#include <stdio.h>
#include <libxml/parser.h>
#include <assert.h>

#include <GDD/arch_xml.h>
#include <GDD/ddg_xml.h>
#include <RS/rs.h>

int main(int argc, char *argv[]) {
  // initialise XML parser
  xmlInitParser();

  // read architecture
  GDD_Arch *arch = GDD_arch_of_xml_file("arch.xml");
  assert(arch != NULL);
      
  // read ddg (ignore interloop dependency)
  SCEDA_Graph *dag = GDD_of_xml_file(TRUE, arch, "ddg.xml",NULL);
  assert(dag != NULL);

  // clean up XML parser
  xmlCleanupParser();

  // first estimate saturation of type GR
  fprintf(stdout,"Approximated register saturation for values of type %s is %d.\n", "GR", RS_estimate_saturation_TC(dag, "GR", TRUE));

  // show a set of saturating values
  SCEDA_List *sat_values;
  if(RS_estimate_saturating_values_TC(dag, "BR", &sat_values) == 0) {
    // first print the size of the list = saturation
    fprintf(stdout,"%d saturating values of type %s.\n",SCEDA_list_size(sat_values), "BR");

    // iterate over the values and print them
    SCEDA_ListIterator vals;
    SCEDA_list_iterator_init(sat_values, &vals);
    while(SCEDA_list_iterator_has_next(&vals)) {
      SCEDA_Vertex *v = SCEDA_list_iterator_next(&vals);
      GDD_opcode_fprint(stdout,GDD_vertex_get_opcode(v));
    }
    SCEDA_list_iterator_cleanup(&vals);

    // delete the list
    SCEDA_list_delete(sat_values);
  }

  // note that if several saturation are to be estimated, it is better to build explicitly the DDAG
  GDD_DAG *ddg = GDD_dag_create(dag); 
  assert(ddg != NULL);

  // iterate over all types of the architecture
  GDD_ArchTypesIterator all_types;
  GDD_arch_types_iterator_init(arch, &all_types);
  while(GDD_arch_types_iterator_has_next(&all_types)) {
    char *type = GDD_arch_types_iterator_next(&all_types);
	
    fprintf(stdout,"Approximated register saturation for values of type %s is %d.\n", type, RS_estimate_saturation(ddg, type,TRUE));
  }
  GDD_arch_types_iterator_cleanup(&all_types);

  GDD_dag_delete(ddg); // clear the DDAG

  // compute optimal saturation of type GR in 10 seconds max
  int timeout_flag;
  int rs = RS_compute_optimal_saturation(dag,"GR",10.0,&timeout_flag);
  if(timeout_flag) {
    fprintf(stderr,"Computation was interrupted.\n");
  }
  if(rs < 0) {
    fprintf(stdout,"Error while computing optimal saturation of type %s.\n","GR");
  } else {
    fprintf(stdout,"Optimal saturation of type %s is %d.\n","GR",rs);
  }

  SCEDA_graph_delete(dag);
  GDD_arch_delete(arch);
      
  return 0;
}
