/*
   Copyright Universite de Versailles Saint-Quentin en Yvelines 2009
   AUTHORS: Sebastien Briais, Sid-Ahmed-Ali Touati

   This file is part of RS.
   
   RS is free software: you can redistribute it and/or modify it
   under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation, either version 3 of the
   License, or (at your option) any later version.
   
   RS is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with RS.  If not, see
   <http://www.gnu.org/licenses/>.
*/
#include <SCEDA/list.h>
#include <SCEDA/graph.h>
#include <SCEDA/graph_dag.h>
#include <SCEDA/common.h>

#include "util.h"

#include "greedy.h"
#include "pkill.h"
#include "dvk.h"

// first chose killers that have no descendant values
#define V0 1
// modified cost function
#define V1 1

typedef struct {
  SCEDA_Vertex *in_pk;
  int visited;
  int succ_taken;
  int pred_taken;
  SCEDA_List *pred;
  SCEDA_List *desc_values;
} State;

static inline State *new_state() {
  State *st = (State *)safe_malloc(sizeof(State));
  st->pred = SCEDA_list_create(NULL);
  st->desc_values = SCEDA_list_create(NULL);
  return st;
}

static void delete_state(State *st) {
  SCEDA_list_delete(st->pred);
  SCEDA_list_delete(st->desc_values);
  free(st);
}

static int saturate_succ(SCEDA_HashMap *v_state, SCEDA_HashSet *s_cb, SCEDA_HashSet *t_cb) {
  int changed = FALSE;
  SCEDA_HashSetIterator s_vertice;
  SCEDA_hashset_iterator_init(s_cb, &s_vertice);
  while(SCEDA_hashset_iterator_has_next(&s_vertice)) {
    SCEDA_Vertex *s = SCEDA_hashset_iterator_next(&s_vertice);
    State *st_s = SCEDA_hashmap_get(v_state, s);
    if(!st_s->succ_taken) {
      st_s->succ_taken = TRUE;
      SCEDA_Vertex *pk_s = st_s->in_pk;
      SCEDA_VertexSuccIterator succ;
      SCEDA_vertex_succ_iterator_init(pk_s, &succ);
      while(SCEDA_vertex_succ_iterator_has_next(&succ)) {
	SCEDA_Vertex *pk_t = SCEDA_vertex_succ_iterator_next(&succ);
	SCEDA_Vertex *t = SCEDA_vertex_get_data(SCEDA_Vertex *, pk_t);
	if(SCEDA_hashset_add(t_cb, t) == 0) {
	  changed = TRUE;
	}
      }
      SCEDA_vertex_succ_iterator_cleanup(&succ);
    }
  }
  SCEDA_hashset_iterator_cleanup(&s_vertice);
  return changed;
}

static int saturate_pred(SCEDA_Graph *pkg, SCEDA_HashMap *v_state, SCEDA_HashSet *s_cb, SCEDA_HashSet *t_cb) {
  int changed = FALSE;
  SCEDA_HashSetIterator t_vertice;
  SCEDA_hashset_iterator_init(t_cb, &t_vertice);
  while(SCEDA_hashset_iterator_has_next(&t_vertice)) {
    SCEDA_Vertex *t = SCEDA_hashset_iterator_next(&t_vertice);
    State *st_t = SCEDA_hashmap_get(v_state, t);
    if(!st_t->pred_taken) {
      st_t->pred_taken = TRUE;
      SCEDA_Vertex *pk_t = st_t->in_pk;
      SCEDA_VertexPredIterator pred;
      SCEDA_vertex_pred_iterator_init(pk_t, &pred);
      while(SCEDA_vertex_pred_iterator_has_next(&pred)) {
	SCEDA_Vertex *pk_s = SCEDA_vertex_pred_iterator_next(&pred);
	SCEDA_Vertex *s = SCEDA_vertex_get_data(SCEDA_Vertex *, pk_s);
	if(SCEDA_hashset_add(s_cb, s) == 0) {
	  changed = TRUE;
	}
      }
      SCEDA_vertex_pred_iterator_cleanup(&pred);
    }
  }
  SCEDA_hashset_iterator_cleanup(&t_vertice);
  return changed;
}

#if(0)
static SCEDA_Vertex *select_random_killer(SCEDA_HashSet *t_cb, SCEDA_HashSet *x, SCEDA_HashSet *y, SCEDA_HashMap *v_state) {
  int n = SCEDA_hashset_size(t_cb);
  int i = n * (((double)rand()) / ((double)RAND_MAX));
  if(i < 0) {
    i = 0;
  } else if(i > n-1) {
    i = n-1;
  } 
  
  SCEDA_Vertex *tmax = NULL;

  SCEDA_HashSetIterator killers;
  SCEDA_hashset_iterator_init(t_cb, &killers);
  while(SCEDA_hashset_iterator_has_next(&killers)) {
    SCEDA_Vertex *t = SCEDA_hashset_iterator_next(&killers);
    if(i == 0) {
      tmax = t;
      break;
    }
    i--;
  }
    
  //  assert(tmax != NULL);
  return tmax;
}
#endif

/** Select a "best" killer in t_cb */
static SCEDA_Vertex *select_killer(SCEDA_HashSet *t_cb, SCEDA_HashSet *x, SCEDA_HashSet *y, SCEDA_HashMap *v_state) {
  SCEDA_Vertex *tmax = NULL;
  int pmax = 0;
  int qmax = 1;

  SCEDA_HashSetIterator children;
  SCEDA_hashset_iterator_init(t_cb, &children);
  while(SCEDA_hashset_iterator_has_next(&children)) {
    SCEDA_Vertex *t = SCEDA_hashset_iterator_next(&children);
    State *st_t = SCEDA_hashmap_get(v_state, t);
   
    // compute cardinal |pred(t) \cap X|
    int p = 0;
    
    SCEDA_ListIterator pred;
    SCEDA_list_iterator_init(st_t->pred, &pred);
    while(SCEDA_list_iterator_has_next(&pred)) {
      SCEDA_Vertex *s = SCEDA_list_iterator_next(&pred);
      if(SCEDA_hashset_contains(x, s)) {
	p++;
      }
    }
    SCEDA_list_iterator_cleanup(&pred);

    // we want to kill at least one value!!
    if(p == 0) {
      continue;
    }

#if(V0)
    if(SCEDA_list_size(st_t->desc_values) == 0) {
      tmax = t;
      pmax = p;
      qmax = 1;
      break;
    } 
#endif
    
    // compute cardinal |desc_values(t) \cup Y|
    int q = SCEDA_hashset_size(y) + SCEDA_list_size(st_t->desc_values);

    // Grassman formula |A \cup B| = |A| + |B| - |A \cap B|
    SCEDA_ListIterator desc;
    SCEDA_list_iterator_init(st_t->desc_values, &desc);
    while(SCEDA_list_iterator_has_next(&desc)) {
      SCEDA_Vertex *v = SCEDA_list_iterator_next(&desc);
      if(SCEDA_hashset_contains(y, v)) {
	q--;
      }
    }
    SCEDA_list_iterator_cleanup(&desc);
    
#if(V1)
    q++;
#else
    q = (q == 0) ? 1 : q; 
#endif

    if(tmax == NULL) {
      tmax = t;
      pmax = p;
      qmax = q;
    } else {
      if(p * qmax > pmax * q) {
	tmax = t;
	pmax = p;
	qmax = q;
      } else if(p * qmax == pmax * q) {
      }
    }
  }
  SCEDA_hashset_iterator_cleanup(&children);
  
  //  assert(tmax != NULL);

  return tmax;
}

/** Compute killing function on given cbc, according to greedy-k
    algorithm */
static void compute_greedy_killing(GDD_DAG *ddg, SCEDA_HashSet *s_cb, SCEDA_HashSet *t_cb, SCEDA_HashMap *v_state, SCEDA_HashMap *killing_map) {
  SCEDA_HashSet *x = s_cb; // shallow copy
  SCEDA_HashSet *y = SCEDA_vertex_set_create();
  
  while(SCEDA_hashset_size(x) > 0) {
    //    SCEDA_Vertex *tmax = select_random_killer(t_cb, x, y, v_state);
    SCEDA_Vertex *tmax = select_killer(t_cb, x, y, v_state);
    if(tmax == NULL) {
      // necessarily, tmax consumes the last nodes
      break;
    } else {
      State *st_tmax = SCEDA_hashmap_get(v_state, tmax);
      
      SCEDA_ListIterator pred;
      SCEDA_list_iterator_init(st_tmax->pred, &pred);
      while(SCEDA_list_iterator_has_next(&pred)) {
	SCEDA_Vertex *s = SCEDA_list_iterator_next(&pred);
	if(!SCEDA_hashmap_contains_key(killing_map, s)) {
	  SCEDA_hashmap_put(killing_map, s, tmax, NULL);
	}
	SCEDA_hashset_remove(x, (void **)&s);
      }
      SCEDA_list_iterator_cleanup(&pred);
    
      SCEDA_ListIterator desc;
      SCEDA_list_iterator_init(st_tmax->desc_values, &desc);
      while(SCEDA_list_iterator_has_next(&desc)) {
	SCEDA_Vertex *v = SCEDA_list_iterator_next(&desc);
	SCEDA_hashset_add(y, v);
      }
      SCEDA_list_iterator_cleanup(&desc);
      
      SCEDA_hashset_remove(t_cb, (void **)&tmax);
    }
  }
  
  SCEDA_hashset_delete(y);
}

/** Compute s_cb and t_cb.

    Compute also the predecessors and the descendant values of
    vertices in t_cb.

    Initially, it is assumed that s_cb consists of one non visited
    value (the seed). */
static void compute_cbc(GDD_DAG *ddg, SCEDA_Graph *pkg, const char *type, SCEDA_HashMap *v_state, SCEDA_HashSet *s_cb, SCEDA_HashSet *t_cb) {
  /** First saturate s_cb and t_cb by applying succ and pred until
      reaching a fixpoint */
  saturate_succ(v_state, s_cb, t_cb);

  int changed = TRUE;
  while(changed) {
    changed = FALSE;
    if(saturate_pred(pkg, v_state, s_cb, t_cb)) {
      changed = saturate_succ(v_state, s_cb, t_cb);
    }
  }
  
  /** Mark all vertices of s_cb as visited */
  SCEDA_HashSetIterator s_vertice;
  SCEDA_hashset_iterator_init(s_cb, &s_vertice);
  while(SCEDA_hashset_iterator_has_next(&s_vertice)) {
    SCEDA_Vertex *s = SCEDA_hashset_iterator_next(&s_vertice);
    State *st_s = SCEDA_hashmap_get(v_state, s);
    st_s->visited = TRUE;
    // the following line is intentionnally commented
    // SCEDA_hashset_remove(t_cb, s); // remove s from t_cb 
  }
  SCEDA_hashset_iterator_cleanup(&s_vertice);

  /** S and T do not have to be disjoint.
      Indeed, think to the graph:
      V = 1,2,3,4,5,6
      E = 1->2, 3->2, 3->4, 5->4, 5->1, 6->1
      then 
      S = 1,3,5,6
      T = 1,2,4
      Ecb = E 
      1 the only possible killer for 6 and those must remain in T */
  
  /** Compute the predecessors of vertices in t_cb */
  SCEDA_hashset_iterator_init(s_cb, &s_vertice);
  while(SCEDA_hashset_iterator_has_next(&s_vertice)) {
    SCEDA_Vertex *s = SCEDA_hashset_iterator_next(&s_vertice);
    State *st_s = SCEDA_hashmap_get(v_state, s);
    SCEDA_Vertex *pk_s = st_s->in_pk;
    SCEDA_VertexSuccIterator succ;
    SCEDA_vertex_succ_iterator_init(pk_s, &succ);
    while(SCEDA_vertex_succ_iterator_has_next(&succ)) {
      SCEDA_Vertex *pk_t = SCEDA_vertex_succ_iterator_next(&succ);
      SCEDA_Vertex *t = SCEDA_vertex_get_data(SCEDA_Vertex *, pk_t);
      if(SCEDA_hashset_contains(t_cb, t)) {
	State *st_t = SCEDA_hashmap_get(v_state, t);
	SCEDA_list_ins_next(st_t->pred, NULL, s);
      }
    }
    SCEDA_vertex_succ_iterator_cleanup(&succ);
  }
  SCEDA_hashset_iterator_cleanup(&s_vertice);
  
  /** Compute the descendant values of vertices in t_cb */
  SCEDA_HashSetIterator t_vertice;
  SCEDA_hashset_iterator_init(t_cb, &t_vertice);
  while(SCEDA_hashset_iterator_has_next(&t_vertice)) {
    SCEDA_Vertex *t = SCEDA_hashset_iterator_next(&t_vertice);
    State *st_t = SCEDA_hashmap_get(v_state, t);
    GDD_DAG_UpsetIterator desc;
    GDD_dag_upset_iterator_init(ddg, t, &desc);
    while(GDD_dag_upset_iterator_has_next(&desc)) {
      SCEDA_Vertex *v = GDD_dag_upset_iterator_next(&desc);
      if(GDD_vertex_is_value_of_type(v, type)) {
	SCEDA_list_ins_next(st_t->desc_values, NULL, v);
      }
    }
    GDD_dag_upset_iterator_cleanup(&desc);
    if(GDD_vertex_is_value_of_type(t, type)) {
      // count t itself when it is a value!!
      SCEDA_list_ins_next(st_t->desc_values, NULL, t); 
    }
  }
  SCEDA_hashset_iterator_cleanup(&t_vertice);
}

/** Compute killing function according to greedy-k algorithm */
static SCEDA_HashMap *RS_ddag_greedy_killing_function(GDD_DAG *ddg, SCEDA_Graph *pkg, const char *type) {
  SCEDA_HashMap *killing_map = SCEDA_vertex_map_create(NULL);
  SCEDA_HashMap *v_state = SCEDA_vertex_map_create((SCEDA_delete_fun)delete_state);
  SCEDA_VerticesIterator vertice;
  SCEDA_vertices_iterator_init(pkg, &vertice);
  while(SCEDA_vertices_iterator_has_next(&vertice)) {
    SCEDA_Vertex *pk_v = SCEDA_vertices_iterator_next(&vertice);
    SCEDA_Vertex *v = SCEDA_vertex_get_data(SCEDA_Vertex *, pk_v);
    State *st_v = new_state();
    st_v->in_pk = pk_v;
    st_v->visited = FALSE;
    st_v->succ_taken = FALSE;
    st_v->pred_taken = FALSE;
    SCEDA_hashmap_put(v_state, v, st_v, NULL);
/*     if(GDD_vertex_is_value_of_type(v, type)) { */
/*       SCEDA_hashmap_put(killing_map, v, RS_ddag_sink(ddg), NULL); */
/*     } */
  }
  SCEDA_vertices_iterator_cleanup(&vertice);

  SCEDA_vertices_iterator_init(pkg, &vertice);
  while(SCEDA_vertices_iterator_has_next(&vertice)) {
    SCEDA_Vertex *pk_v = SCEDA_vertices_iterator_next(&vertice);
    SCEDA_Vertex *v = SCEDA_vertex_get_data(SCEDA_Vertex *, pk_v);
    if(GDD_vertex_is_value_of_type(v, type)) {
      State *st_v = SCEDA_hashmap_get(v_state, v);
      if(!st_v->visited) {
	SCEDA_HashSet *s_cb = SCEDA_vertex_set_create();
	SCEDA_HashSet *t_cb = SCEDA_vertex_set_create();

	SCEDA_hashset_add(s_cb, v);
	compute_cbc(ddg, pkg, type, v_state, s_cb, t_cb);

	compute_greedy_killing(ddg, s_cb, t_cb, v_state, killing_map);

	SCEDA_hashset_delete(s_cb);
	SCEDA_hashset_delete(t_cb);
      }
    }
  }
  SCEDA_vertices_iterator_cleanup(&vertice);

  SCEDA_hashmap_delete(v_state);

  return killing_map;
}

/** Compute a killing function according to Greedy-k algorithm.

    @param[in] ddg = GDD_DAG
    @param[in] type = considered type

    @return a killing function */
SCEDA_HashMap *RS_ddag_greedy_k(GDD_DAG *ddg, const char *type) {
  SCEDA_Graph *pkg = RS_ddag_pkill(ddg, type);

  SCEDA_HashMap *killing = RS_ddag_greedy_killing_function(ddg, pkg, type);

  SCEDA_graph_delete(pkg);

  return killing;
}

/** Return an approximated set of saturating values for given type.
    Use Greedy-k algorithm.

    @param[in] ddg = GDD_DAG
    @param[in] type = considered type

    @return a list of saturating values */
SCEDA_List *RS_ddag_estimate_saturating_values(GDD_DAG *ddg, const char *type) {
  SCEDA_Graph *pkg = RS_ddag_pkill(ddg, type);
  SCEDA_HashMap *killing = RS_ddag_greedy_killing_function(ddg, pkg, type);

  SCEDA_List *amk = RS_ddag_saturating_values(ddg, type, pkg, killing); 

  SCEDA_hashmap_delete(killing);

  SCEDA_graph_delete(pkg);

  return amk;
}

/** Estimate the register saturation for given type.
    Use Greedy-k algorithm.

    @param[in] ddg = GDD_DAG
    @param[in] type = considered type
    @param[in] fast = TRUE to use fast version (variant 2 heuristic), FALSE to use precise version (variant 1 heuristic)

    @return an approximation of the register saturation */
int RS_ddag_estimate_saturation(GDD_DAG *ddg, const char *type, int fast) {
  SCEDA_Graph *pkg = RS_ddag_pkill(ddg, type);

  SCEDA_HashMap *killing = RS_ddag_greedy_killing_function(ddg, pkg, type);

  int n = RS_ddag_register_need(ddg, type, pkg, killing, fast);

  SCEDA_hashmap_delete(killing);

  SCEDA_graph_delete(pkg);

  return n;
}

