/*
   Copyright Universite de Versailles Saint-Quentin en Yvelines 2009
   AUTHORS: Sebastien Briais, Sid-Ahmed-Ali Touati

   This file is part of SIRA.
   
   SIRA is free software: you can redistribute it and/or modify it
   under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation, either version 3 of the
   License, or (at your option) any later version.
   
   SIRA is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with SIRA.  If not, see
   <http://www.gnu.org/licenses/>.
*/
#include "isolver.h"
#include "solver.h"
#include "util.h"

#include "weightmap.h"

/* Include string match and hash functions */
#include "stringutil.h"

#include "boxed.h"

/** Undo the modifications performed by the sequential siralina solver.

    @param[in] loop = siraloop
    @param[in] seq_solution = sequential siralina solution to undo

    Note that the hashmap is cleared but not deleted. */
void SIRA_isolver_undo(SIRA_Loop *loop, SCEDA_HashMap *seq_solution) {
  SCEDA_HashMapIterator sol;
  SCEDA_hashmap_iterator_init(seq_solution, &sol);
  while(SCEDA_hashmap_iterator_has_next(&sol)) {
    char *type;
    SCEDA_List *antidep_edges = SCEDA_hashmap_iterator_next(&sol, &type);
    GDD_delete_edges(SIRA_loop_ddg(loop), antidep_edges);
  }
  SCEDA_hashmap_iterator_cleanup(&sol);

  SIRA_loop_update(loop);

  SCEDA_hashmap_clear(seq_solution);
}

/** Check whether the solution satisfies the register constraints.

    @param[in] regconstraints_map = register constraints
    @param[in] costmap = map (type -> boxed(int))

    @return TRUE iff constraints are satisfied */
static int SIRA_isolver_is_solution(SIRA_RegConstraintsMap *regconstraints_map, SCEDA_HashMap *costmap) {
  int is_solution = TRUE;
  //
  SCEDA_HashMapIterator costs;
  SCEDA_hashmap_iterator_init(costmap, &costs);
  while(SCEDA_hashmap_iterator_has_next(&costs)) {
    char *type;
    boxed(int) cost = SCEDA_hashmap_iterator_next(&costs, &type);
    int max = SIRA_regconstraints_map_get_constraint_of_type(regconstraints_map, type);
    if(boxed_get(cost) > max) {
      is_solution = FALSE;
      break;
    }
  }
  SCEDA_hashmap_iterator_cleanup(&costs);
  //
  return is_solution;
}

/** Check whether the solution is valid, i.e. if the DDG is
    lexicographic positive.

    @param[in] loop = SiraLoop DDG
    @param[in] settings = solver settings

    @return TRUE iff the DDG is lexicographic positive 
    note that it is trivially TRUE when UAL semantics is enabled */
static int SIRA_isolver_solution_is_lexicographic_positive(SIRA_Loop *loop, SIRA_Param *settings) {
  int result = TRUE;
  if(SIRA_param_get_ual_flag(settings)) {
    return result;
  }
  SCEDA_List *cycle;
  if(!GDD_is_lexicographic_positive(SIRA_loop_ddg(loop), &cycle)) {
    SCEDA_list_delete(cycle);
    result = FALSE;
  }
  return result;
}

/** Apply sequential siralina heuristic for given II.

    @param[in] loop = siraloop
    @param[in] settings = solver settings
    @param[in] types = list of types to process in specified order
    @param[in] II = II (if negative, fails)
    @param[out] seq_solution = computed solution (map from types to antidep edges list)
    @param[out] costmap = cost of solution (map from types to integer pointers)

    @return 0 if a solution was found, otherwise -1 
    Beware that the loop is modified as a side-effect when a solution is found. */
static int SIRA_isolver_single_internal(SIRA_Loop *loop, SIRA_Param *settings, SCEDA_List *types, int II, SCEDA_HashMap **solution, SCEDA_HashMap **costmap) {
  if(II < 0) {
    return -1;
  }

  int is_solution = TRUE;
  SCEDA_HashMap *result = SCEDA_hashmap_create((SCEDA_delete_fun)free, (SCEDA_delete_fun)SCEDA_list_delete, (SCEDA_match_fun)string_match, (SCEDA_hash_fun)string_hash);
  SCEDA_HashMap *allcosts = SCEDA_hashmap_create((SCEDA_delete_fun)free, (SCEDA_delete_fun)boxed_delete, (SCEDA_match_fun)string_match, (SCEDA_hash_fun)string_hash);

  SIRA_WeightMap *weight_map = SIRA_weight_map_create();

  SCEDA_ListIterator all_types;
  SCEDA_list_iterator_init(types, &all_types);
  while(SCEDA_list_iterator_has_next(&all_types)) {
    char *type = SCEDA_list_iterator_next(&all_types);

    if(SCEDA_hashmap_contains_key(result, type)) {
      continue;
    }
    
    SIRA_weight_map_clear(weight_map);
    SIRA_weight_map_set_weight_of_type(weight_map, type, 1);

    SCEDA_HashMap *reuse_decisions;
    if(SIRA_solver_single(loop, settings, weight_map, II, &reuse_decisions) == 0) {
      SCEDA_List *reuse_edges = SCEDA_hashmap_get(reuse_decisions, type);
      safe_ptr(reuse_edges);
      int cost = SIRA_reuse_decisions_get_cost(reuse_edges);
      SCEDA_List *antidep_edges = SIRA_reuse_apply_decisions(SIRA_loop_ddg(loop), type, reuse_edges, SIRA_param_get_ual_flag(settings));
      SCEDA_hashmap_put(result, safe_strdup(type), antidep_edges, NULL);
      SCEDA_hashmap_put(allcosts, safe_strdup(type), boxed_create(int, cost), NULL);
      SCEDA_hashmap_delete(reuse_decisions);

      if(SIRA_loop_update(loop) != 0) {
	is_solution = FALSE;
	break;
      }
    } else {
      is_solution = FALSE;
      break;
    }
  }
  SCEDA_list_iterator_cleanup(&all_types);

  SIRA_weight_map_delete(weight_map);

  if(is_solution) {
    *solution = result;
    *costmap = allcosts;
    return 0;
  } else {
    SIRA_isolver_undo(loop, result);
    SCEDA_hashmap_delete(result);
    SCEDA_hashmap_delete(allcosts);
    return -1;
  }
}

/** Apply sequential siralina heuristic for given II.

    It will switch to UAL semantics if no valid solution can be found
    with NUAL semantics.

    @param[in] loop = siraloop
    @param[in] settings = solver settings (semantics might be modified)
    @param[in] types = list of types to process in specified order
    @param[in] II = II (if negative, fails)
    @param[out] seq_solution = computed solution (map from types to antidep edges list)
    @param[out] costmap = cost of solution (map from types to integer pointers)

    @return 0 if a solution was found, otherwise -1 
    Beware that the loop is modified as a side-effect when a solution is found. */
int SIRA_isolver_single(SIRA_Loop *loop, SIRA_Param *settings, SCEDA_List *types, int II, SCEDA_HashMap **solution, SCEDA_HashMap **costmap) {
  if(SIRA_isolver_single_internal(loop, settings, types, II, solution, costmap) == 0) {
    if(!SIRA_param_get_check_flag(settings)) {
      return 0;
    }
    if(SIRA_param_get_ual_flag(settings)) {
      /* if UAL semantics then no need to check solution */
      return 0;
    } else {
      if(SIRA_isolver_solution_is_lexicographic_positive(loop, settings)) {
	return 0;
      }
      /* the solution was not lexicographic positive */
      /* retry with UAL semantics */
      SIRA_isolver_undo(loop, *solution);
      SCEDA_hashmap_delete(*solution);
      SCEDA_hashmap_delete(*costmap);
      SIRA_param_set_ual_flag(settings, TRUE);
      return SIRA_isolver_single_internal(loop, settings, types, II, solution, costmap);
    }
  } else {
    if(!SIRA_param_get_check_flag(settings)) {
      return -1;
    }
    if(!SIRA_param_get_ual_flag(settings)) {
      SIRA_param_set_ual_flag(settings, TRUE);
      return SIRA_isolver_single_internal(loop, settings, types, II, solution, costmap);      
    }
  }
  return -1;
}

/** Apply sequential siralina heuristic for every II between IImin and
    IImax and stops when a solution is found (ie satisfies the
    register constraints).

    @param[in] loop = siraloop
    @param[in] regconstraints_map = register constraints
    @param[in] settings = solver settings
    @param[in] types = list of types to process in specified order
    @param[in] IImin = IImin (if negative, IImin of the loop will be used)
    @param[in] IImax = IImax (if negative, IImax of the loop will be used) 
    @param[out] seq_solution = computed solution (map from types to antidep edges list)
    @param[out] costmap = cost of solution (map from types to integer pointers)

    @return II >= 0 if a solution was found, otherwise -1 
    Beware that the loop is modified as a side-effect when a solution is found. */
static int SIRA_isolver_search_internal(SIRA_Loop *loop, SIRA_RegConstraintsMap *regconstraints_map, SIRA_Param *settings, SCEDA_List *types, int IImin, int IImax, SCEDA_HashMap **seq_solution, SCEDA_HashMap **costmap) {
  int found = FALSE;

  if(IImin < 0) {
    IImin = SIRA_loop_IImin(loop);
  }
  if(IImax < 0) {
    IImax = SIRA_loop_IImax(loop);
  }
  if(IImax < 0) {
    return -1;
  }
  if(IImax < IImin) {
    return -1;
  }

  SCEDA_HashMap *result;
  SCEDA_HashMap *allcosts;
 
  int II;

  II = IImin;
  if(SIRA_isolver_single_internal(loop, settings, types, II, &result, &allcosts) == 0) {
    if(SIRA_isolver_is_solution(regconstraints_map, allcosts)) {
      found = TRUE;
    } else {
      SIRA_isolver_undo(loop, result);
      SCEDA_hashmap_delete(result);
      SCEDA_hashmap_delete(allcosts);
    }
  }
  
  if(!found) {
    II = IImax;
    if(SIRA_isolver_single_internal(loop, settings, types, II, &result, &allcosts) == 0) {
      found = SIRA_isolver_is_solution(regconstraints_map, allcosts);
      SIRA_isolver_undo(loop, result);
      SCEDA_hashmap_delete(result);
      SCEDA_hashmap_delete(allcosts);
    }

    if(found) {
      found = FALSE;
      for(II = IImin+1; II <= IImax; II++) {
	if(SIRA_isolver_single_internal(loop, settings, types, II, &result, &allcosts) == 0) {
	  if(SIRA_isolver_is_solution(regconstraints_map, allcosts)) {
	    found = TRUE;
	    break;
	  } else {
	    SIRA_isolver_undo(loop, result);
	    SCEDA_hashmap_delete(result);
	    SCEDA_hashmap_delete(allcosts);
	  }
	}
      }
    }
  }

  if(found) {
    *seq_solution = result;
    *costmap = allcosts;
    return II;
  } else {
    return -1;
  }
}

/** Apply sequential siralina heuristic for every II between IImin and
    IImax and stops when a solution is found (ie satisfies the
    register constraints).

    @param[in] loop = siraloop
    @param[in] regconstraints_map = register constraints
    @param[in] settings = solver settings
    @param[in] types = list of types to process in specified order
    @param[in] IImin = IImin (if negative, IImin of the loop will be used)
    @param[in] IImax = IImax (if negative, IImax of the loop will be used) 
    @param[out] seq_solution = computed solution (map from types to antidep edges list)
    @param[out] costmap = cost of solution (map from types to integer pointers)

    @return II >= 0 if a solution was found, otherwise -1 
    Beware that the loop is modified as a side-effect when a solution is found. */
int SIRA_isolver_search(SIRA_Loop *loop, SIRA_RegConstraintsMap *regconstraints_map, SIRA_Param *settings, SCEDA_List *types, int IImin, int IImax, SCEDA_HashMap **solution, SCEDA_HashMap **costmap) {
  int check_flag = SIRA_param_get_check_flag(settings);
  SIRA_param_set_check_flag(settings, FALSE);
  int II = SIRA_isolver_search_internal(loop, regconstraints_map, settings, types, IImin, IImax, solution, costmap);
  if(II >= 0) {
    if(!check_flag) {
      return II;
    }
    if(SIRA_param_get_ual_flag(settings)) {
      SIRA_param_set_check_flag(settings, check_flag);
      return II;
    }
    if(SIRA_isolver_solution_is_lexicographic_positive(loop, settings)) { 
      SIRA_param_set_check_flag(settings, check_flag);
      return II;
    }
    SIRA_isolver_undo(loop, *solution);
    SCEDA_hashmap_delete(*solution);
    SCEDA_hashmap_delete(*costmap);
    SIRA_param_set_ual_flag(settings, TRUE);
    II = SIRA_isolver_search_internal(loop, regconstraints_map, settings, types, IImin, IImax, solution, costmap);
    SIRA_param_set_check_flag(settings, check_flag);
    return II;
  } else {
    if(!check_flag) {
      return II;
    }
    if(!SIRA_param_get_ual_flag(settings)) {
      SIRA_param_set_ual_flag(settings, TRUE);
      II = SIRA_isolver_search_internal(loop, regconstraints_map, settings, types, IImin, IImax, solution, costmap);
      SIRA_param_set_check_flag(settings, check_flag);
      return II;
    }
  }
  return II;
}

/** Apply sequential siralina heuristic for every II between IImin and IImax and
    print the results (used for statistics).

    @param[in] loop = siraloop
    @param[in] settings = solver settings
    @param[in] types = list of types to process in specified order
    @param[in] IImin = IImin (if negative, IImin of the loop will be used)
    @param[in] IImax = IImax (if negative, IImax of the loop will be used) */
void SIRA_isolver_stats(SIRA_Loop *loop, SIRA_Param *settings, SCEDA_List *types, int IImin, int IImax) {
  if(IImin < 0) {
    IImin = SIRA_loop_IImin(loop);
  }
  if(IImax < 0) {
    IImax = SIRA_loop_IImax(loop);
  }

  int II;
  for(II = IImin; II <= IImax; II++) {
    SCEDA_HashMap *result;
    SCEDA_HashMap *allcosts;
    if(SIRA_isolver_single_internal(loop, settings, types, II, &result, &allcosts) == 0) {
      fprintf(stdout,"II= %d\tL= %d\t", II, SIRA_loop_get_L(loop, II));

      SCEDA_HashMapIterator costs;
      SCEDA_hashmap_iterator_init(allcosts, &costs);
      while(SCEDA_hashmap_iterator_has_next(&costs)) {
	char *type;
	boxed(int) cost = SCEDA_hashmap_iterator_next(&costs, &type);
	fprintf(stdout, "R_%s= %d\t", type, boxed_get(cost));
      }
      SCEDA_hashmap_iterator_cleanup(&costs);

      fprintf(stdout,"\n");

      SIRA_isolver_undo(loop, result);
      SCEDA_hashmap_delete(result);
      SCEDA_hashmap_delete(allcosts);
    }
  }
}
