/*
   Copyright Universite de Versailles Saint-Quentin en Yvelines 2009
   AUTHORS: Sebastien Briais, Sid-Ahmed-Ali Touati

   This file is part of SIRA.
   
   SIRA is free software: you can redistribute it and/or modify it
   under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation, either version 3 of the
   License, or (at your option) any later version.
   
   SIRA is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with SIRA.  If not, see
   <http://www.gnu.org/licenses/>.
*/
#include <string.h>
#include <limits.h>
#include <SCEDA/graph_dag.h>
#include <SCEDA/graph_path.h>

#include "loop.h"
#include "util.h"
#include <GDD/opcode.h>

#include <stdio.h>

/** Compute a bound for L, it is simply the sum of all the edhe
    latencies 

    @param[in] g = DDG 

    return Lmax */
static int SIRA_loop_compute_Lmax(SCEDA_Graph *g) {
  int L = 0;

  SCEDA_VerticesIterator vertices;
  SCEDA_vertices_iterator_init(g, &vertices);
  while(SCEDA_vertices_iterator_has_next(&vertices)) {
    SCEDA_Vertex *v = SCEDA_vertices_iterator_next(&vertices);

    SCEDA_OutEdgesIterator edges;
    SCEDA_out_edges_iterator_init(v, &edges);
    if(SCEDA_out_edges_iterator_has_next(&edges)) {
      int max_latency = 1;
      while(SCEDA_out_edges_iterator_has_next(&edges)) {
	SCEDA_Edge *e = SCEDA_out_edges_iterator_next(&edges);
	int latency = GDD_edge_latency(e);
	if(latency > max_latency) {
	  max_latency = latency;
	}
      }
      L += max_latency;
    } else {
      L += GDD_vertex_latency(v);
    }
    SCEDA_out_edges_iterator_cleanup(&edges);
  }
  SCEDA_vertices_iterator_cleanup(&vertices);

  return L;
}

/** Compute the maximal distance in the DDG 

    @param[in] g = DDG

    return Lmax */
static int SIRA_loop_compute_max_lambda(SCEDA_Graph *g) {
  int lambda = INT_MIN;

  SCEDA_EdgesIterator edges;
  SCEDA_edges_iterator_init(g, &edges);
  while(SCEDA_edges_iterator_has_next(&edges)) {
    SCEDA_Edge *e = SCEDA_edges_iterator_next(&edges);
    
    if(GDD_edge_distance(e) > lambda) {
      lambda = GDD_edge_distance(e);
    }
  }
  SCEDA_edges_iterator_cleanup(&edges);

  return lambda;
}

/** Returns the duration for given II (depends on the chosen method).

    @param[in] loop = siraloop
    @param[in] II = II

    @return the duration */
int SIRA_loop_get_L(SIRA_Loop *loop, int II) {
  switch(loop->Lmeth) {
  case Lfix: {
    return loop->L;
    break;
  }
  case Lsira: {
    if(loop->Lvals == NULL) {
      int size = loop->IImax - loop->IImin + 1;
      assert(size > 0);
      loop->Lvals = safe_calloc(sizeof(int), size);
      {
	int Lprev = loop->L;
	loop->Lvals[0] = Lprev;
	int i;
	for(i = 1; i < size; i++) {
	  loop->Lvals[i] = Lprev + 1 + (Lprev / (loop->IImin + i));
	  Lprev = loop->Lvals[i];
	}
      }
    } 
    if((loop->IImin <= II) && (II <= loop->IImax)) {
      return loop->Lvals[II-loop->IImin];
    } else {
      return loop->L;
    }
    break;
  }
  case Llao: {
    return (1 << loop->lao_pipelining_degree) * II;
    break;
  }
  default: {
    return loop->L;
    break;
  }
  }
}

/** Free and invalidate precomputed data 

    @param[in] loop = SiraLoop DDG */
static void SIRA_loop_invalid_precalc(SIRA_Loop *loop) {
  if(loop->Lvals != NULL) {
    free(loop->Lvals);
    loop->Lvals = NULL;
  }
}

/** Set min and max bounds for II

    @param[in] loop = SiraLoop DDG
    @param[in] IImin = min II (if <0 then take MII)
    @param[in] IImax = max II (if <0 then take Lmax) */
static void SIRA_loop_set_II_bounds(SIRA_Loop *loop, int IImin, int IImax) {
  SIRA_loop_invalid_precalc(loop);
  if(IImin < 0) {
    IImin = loop->IIrec;
  }
  if(IImax < 0) {
    IImax = loop->Lmax;
  }
  assert(IImin <= IImax);
  loop->IImin = IImin;
  loop->IImax = IImax;
  loop->L = loop->IImax;
}

/** Set the parameter L used by several methods to compute the max
    latency.

    @param[in] loop = siraloop
    @param[in] L = value for the parameter */
void SIRA_loop_set_L(SIRA_Loop *loop, int L) {
  SIRA_loop_invalid_precalc(loop);
  if(L < 0) {
    loop->L = loop->Lmax;
  } else {
    loop->L = L;
  }
  loop->IImax = loop->L;
}

/** Switch to lao method for computing max latency.

    @param[in] loop = siraloop
    @param[in] pipelining_degree = pipelining degree */
void SIRA_loop_set_lao_method(SIRA_Loop *loop, int pipelining_degree) {
  SIRA_loop_invalid_precalc(loop);
  loop->Lmeth = Llao;
  loop->lao_pipelining_degree = pipelining_degree;
  assert(loop->lao_pipelining_degree > 0);
}

/** Switch to sira method for computing max latency.

    @param[in] loop = siraloop */
void SIRA_loop_set_sira_method(SIRA_Loop *loop) {
  SIRA_loop_invalid_precalc(loop);
  loop->Lmeth = Lsira;
}

/** Switch to fixed method for computing max latency.

    @param[in] loop = siraloop */
void SIRA_loop_set_fixed_method(SIRA_Loop *loop) {
  SIRA_loop_invalid_precalc(loop);
  loop->Lmeth = Lfix;
}

static int opp_latency(SCEDA_Edge *e, void *data) {
  return -GDD_edge_latency(e);
}

/** Compute the DAG corresponding to the body of the loop

    @param loop = SiraLoop DDG

    @return 0 in case of success, -1 otherwise (e.g. when body is not a DAG) */
static int SIRA_loop_create_body(SIRA_Loop *loop) {
  if(loop->dag != NULL) {
    return 0;
  }
  SCEDA_Graph *ddg = SIRA_loop_ddg(loop);
  // initialise loop body & related stuff
  SCEDA_Graph *dag = SCEDA_graph_create(NULL, NULL);
  SCEDA_HashMap *ddg_to_body = SCEDA_vertex_map_create(NULL);
  // duplicate vertices
  {
    SCEDA_VerticesIterator vertices;
    SCEDA_vertices_iterator_init(ddg, &vertices);
    while(SCEDA_vertices_iterator_has_next(&vertices)) {
      SCEDA_Vertex *v = SCEDA_vertices_iterator_next(&vertices);
      SCEDA_Vertex *vl = SCEDA_graph_add_vertex(dag, GDD_vertex_get_node_label(v));
      SCEDA_hashmap_put(ddg_to_body, v, vl, NULL);
    }
    SCEDA_vertices_iterator_cleanup(&vertices);
  }
  // duplicate edges
  {
    SCEDA_EdgesIterator edges;
    SCEDA_edges_iterator_init(ddg, &edges);
    while(SCEDA_edges_iterator_has_next(&edges)) {
      SCEDA_Edge *e = SCEDA_edges_iterator_next(&edges);
      if(GDD_edge_distance(e) == 0) {
	SCEDA_Vertex *v_s = SCEDA_edge_source(e);
	SCEDA_Vertex *v_t = SCEDA_edge_target(e);
	
	SCEDA_Vertex *vl_s = SCEDA_hashmap_get(ddg_to_body, v_s);
	SCEDA_Vertex *vl_t = SCEDA_hashmap_get(ddg_to_body, v_t);
	
	SCEDA_graph_add_edge(dag, vl_s, vl_t, GDD_edge_get_edge_label(e));
      }
    }
    SCEDA_edges_iterator_cleanup(&edges);
  }
  // add source & sink
  GDD_Opcode *dummy_op = GDD_opcode_create("dummy_op", 0, 0);
  // this is not correct because dummy_node will label two different vertices
  GDD_NodeLabel *dummy_node = GDD_node_label_create(-1, dummy_op); 
  SCEDA_Vertex *v_source = SCEDA_graph_add_vertex(dag, dummy_node);
  SCEDA_Vertex *v_sink = SCEDA_graph_add_vertex(dag, dummy_node);
  SCEDA_List *dag_edges = SCEDA_list_create(NULL);
  {
    SCEDA_VerticesIterator g_vertice;
    SCEDA_vertices_iterator_init(dag, &g_vertice);
    while(SCEDA_vertices_iterator_has_next(&g_vertice)) {
      SCEDA_Vertex *v = SCEDA_vertices_iterator_next(&g_vertice);
      if(v != v_source) {
	if(SCEDA_vertex_in_deg(v) == 0) {
	  GDD_EdgeLabel *e = GDD_edge_label_create(NULL, "virtual", 0, 0);
	  SCEDA_Edge *virt_e = SCEDA_graph_add_edge(dag, v_source, v, e);
	  safe_call(SCEDA_list_add(dag_edges, virt_e));
	}
      }
      if(v != v_sink) {
	if(SCEDA_vertex_out_deg(v) == 0) {
	  // GDD_EdgeLabel *e = GDD_edge_label_create(NULL, "virtual", GDD_vertex_latency(v), 0);
	  GDD_EdgeLabel *e = GDD_edge_label_create(NULL, "virtual", 0, 0);
	  SCEDA_Edge *virt_e = SCEDA_graph_add_edge(dag, v, v_sink, e);
	  safe_call(SCEDA_list_add(dag_edges, virt_e));
	}
      }
    }
    SCEDA_vertices_iterator_cleanup(&g_vertice);
  }
  //
  loop->dag = dag;
  loop->dag_source = v_source;
  loop->dag_sink = v_sink;
  loop->dummy_op = dummy_op;
  loop->dummy_node = dummy_node;
  loop->dag_edges = dag_edges;
  loop->ddg_to_body = ddg_to_body;
  //
  // prerequisite before computing paths
  if(SCEDA_graph_compute_topological_order(dag) == 0) {
    loop->longest_path_from_source = SCEDA_graph_shortest_path_from_in_dag(dag, v_source, (SCEDA_int_edge_fun)opp_latency, NULL);
    loop->longest_path_to_sink = SCEDA_graph_shortest_path_to_in_dag(dag, v_sink, (SCEDA_int_edge_fun)opp_latency, NULL);
    return 0;
  } else {
    loop->longest_path_from_source = SCEDA_vertex_map_create(NULL);
    loop->longest_path_to_sink = SCEDA_vertex_map_create(NULL);
    return -1; // ERROR, body is not a DAG!
  }
}

/** Delete and invalidate the precomputed body

    @param[in] loop = SiraLoop DDG */
static void SIRA_loop_delete_body(SIRA_Loop *loop) {
  if(loop->dag == NULL) {
    return;
  }
  // delete loop body related stuff
  SCEDA_hashmap_delete(loop->ddg_to_body);
  // delete paths
  SCEDA_hashmap_delete(loop->longest_path_from_source);
  SCEDA_hashmap_delete(loop->longest_path_to_sink);
  // remove source & sink
  {
    GDD_delete_edges(loop->dag, loop->dag_edges);
    SCEDA_list_delete(loop->dag_edges);

    GDD_NodeLabel *node;
    safe_call(SCEDA_graph_remove_vertex(loop->dag, loop->dag_source, (void **)&node));
    safe_call(SCEDA_graph_remove_vertex(loop->dag, loop->dag_sink, (void **)&node));
    GDD_node_label_delete(loop->dummy_node);
    GDD_opcode_delete(loop->dummy_op);
  }
  //
  SCEDA_graph_delete(loop->dag);
  //
  loop->dag = NULL;
  loop->dag_source = NULL;
  loop->dag_sink = NULL;
  loop->dummy_op = NULL;
  loop->dummy_node = NULL;
  loop->dag_edges = NULL;
  loop->ddg_to_body = NULL;
  //
  loop->longest_path_from_source = NULL;
  loop->longest_path_to_sink = NULL;
}

/** Updates the precomputed values. This should be called if the DDG
    structure change. 

    @param[in] loop = siraloop 

    @return 0 in case of success, -1 otherwise */
int SIRA_loop_update(SIRA_Loop *loop) {
  loop->Lmax = SIRA_loop_compute_Lmax(SIRA_loop_ddg(loop));
  loop->maxLambda = SIRA_loop_compute_max_lambda(SIRA_loop_ddg(loop));
  // L values
  SIRA_loop_invalid_precalc(loop);
  // body DAG
  SIRA_loop_delete_body(loop);
  return SIRA_loop_create_body(loop);
}

/** Creates a siraloop.

    @param[in] g = DDG (it is assumed to be lexicographic positive)
    @param[in] IIrec = MII (resource constraints) 

    @return a siraloop. Beware that some values are precomputed using
    the current DDG. If the DDG changes, the values need to be
    updated. */
SIRA_Loop *SIRA_loop_create(SCEDA_Graph *ddg, int IIrec) {
  SIRA_Loop *loop = safe_malloc(sizeof(SIRA_Loop));
  loop->ddg = ddg;
  if(IIrec < 0) {
    IIrec = GDD_MII(ddg);
  }
  loop->IIrec = IIrec;
  // set buffer address to null
  // L values
  loop->Lvals = NULL;
  // body DAG
  loop->dag = NULL;
  loop->dag_source = NULL;
  loop->dag_sink = NULL;
  loop->dummy_op = NULL;
  loop->dag_edges = NULL;
  loop->ddg_to_body = NULL;
  loop->longest_path_from_source = NULL;
  loop->longest_path_to_sink = NULL;
  // update precomputed values
  SIRA_loop_update(loop);
  // default parameters
  SIRA_loop_set_II_bounds(loop, IIrec, loop->Lmax);
  SIRA_loop_set_L(loop, loop->Lmax);
  SIRA_loop_set_fixed_method(loop);
  //
  return loop;
}

/** Deletes a siraloop.

    @param[in] loop = siraloop to delete */
void SIRA_loop_delete(SIRA_Loop *loop) {
  SIRA_loop_invalid_precalc(loop);
  SIRA_loop_delete_body(loop);
  memset(loop, 0, sizeof(SIRA_Loop));
  free(loop);
}

/** Returns the lower bound for scheduling DDG operation u, 
    which must belong to the loop.

    @param[in] loop = siraloop
    @param[in] u = operation

    @return the lower bound */
int SIRA_loop_sigma_min(SIRA_Loop *loop, SCEDA_Vertex *u) {
  SCEDA_Vertex *ul = SCEDA_hashmap_get(loop->ddg_to_body, u);
  SCEDA_PathInfo *info = SCEDA_hashmap_get(loop->longest_path_from_source, ul);
  if(info == NULL) {
    return 0;
  }
  int lp = -info->distance;
  return lp;
}

/** Returns the upper bound for scheduling DDG operation u, 
    which must belong to the loop.

    @param[in] loop = siraloop
    @param[in] u = operation
    @param[in] L = duration

    @return the upper bound */
int SIRA_loop_sigma_max(SIRA_Loop *loop, SCEDA_Vertex *u, int L) {
  SCEDA_Vertex *ul = SCEDA_hashmap_get(loop->ddg_to_body, u);
  SCEDA_PathInfo *info = SCEDA_hashmap_get(loop->longest_path_to_sink, ul);
  if(info == NULL) {
    return 0;
  }
  int lp = -info->distance;
  return L-lp;
}

/** Returns the lower bound for scheduling killer of DDG operation u,
    which must belong to the loop.

    @param[in] loop = siraloop
    @param[in] u = operation
    @param[in] type = killer type
    @param[in] II = initiation interval

    @return the lower bound */
int SIRA_loop_k_min(SIRA_Loop *loop, SCEDA_Vertex *u, const char *type, int II) {
  int k = 0;
  
  SCEDA_HashSet *cons = GDD_consumers_of_type(u, type);

  SCEDA_HashSetIterator consumers;
  SCEDA_hashset_iterator_init(cons, &consumers);
  while(SCEDA_hashset_iterator_has_next(&consumers)) {
    SCEDA_Vertex *consu = SCEDA_hashset_iterator_next(&consumers);
    int ku = SIRA_loop_sigma_min(loop, consu) + GDD_max_distance_flowdep(u, consu, type) * II;
    if(ku > k) {
      k = ku;
    }
  }
  SCEDA_hashset_iterator_cleanup(&consumers);
  
  SCEDA_hashset_delete(cons);

  return k;
}

/** Returns the upper bound for scheduling killer of DDG operation u,
    which must belong to the loop.

    @param[in] loop = siraloop
    @param[in] u = operation
    @param[in] L = duration
    @param[in] type = killer type
    @param[in] II = initiation interval

    @return the lower bound */
int SIRA_loop_k_max(SIRA_Loop *loop, SCEDA_Vertex *u, int L, const char *type, int II) {
  int k = 0;

  SCEDA_HashSet *cons = GDD_consumers_of_type(u, type);

  if(SCEDA_hashset_is_empty(cons)) {
    k = L + SIRA_loop_max_lambda(loop) * II;
  } else {
    SCEDA_HashSetIterator consumers;
    SCEDA_hashset_iterator_init(cons, &consumers);
    while(SCEDA_hashset_iterator_has_next(&consumers)) {
      SCEDA_Vertex *consu = SCEDA_hashset_iterator_next(&consumers);
      int ku = SIRA_loop_sigma_max(loop, consu, L) + GDD_max_distance_flowdep(u, consu, type) * II;
      if(ku > k) {
	k = ku;
      }
    }
    SCEDA_hashset_iterator_cleanup(&consumers);
  }

  SCEDA_hashset_delete(cons);

  return k;
}
