/*
   Copyright Universite de Versailles Saint-Quentin en Yvelines 2009
   AUTHORS: Sebastien Briais, Sid-Ahmed-Ali Touati

   This file is part of SIRA.
   
   SIRA is free software: you can redistribute it and/or modify it
   under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation, either version 3 of the
   License, or (at your option) any later version.
   
   SIRA is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with SIRA.  If not, see
   <http://www.gnu.org/licenses/>.
*/
#ifndef __SIRA_LOOP_H
#define __SIRA_LOOP_H
/** \file loop.h
    \brief Auxiliary data structure used by siralina solver.
    Basically, a SIRA Loop is a DDG + some scheduling information
 */

#include <SCEDA/graph.h>
#include <SCEDA/hashmap.h>

#include <GDD/ddg.h>

typedef enum {
  Lfix,
  Lsira,
  Llao
} SIRA_LMethod;

/** Type of SIRA_Loop */
typedef struct {
  // given parameters
  SCEDA_Graph *ddg;
  int IIrec;
  // computed Lmax
  int Lmax;
  // max lambda
  int maxLambda;
  // II bounds
  int IImin;
  int IImax;
  // L method parameters and precalc
  SIRA_LMethod Lmeth;
  int L;
  int lao_pipelining_degree;
  int *Lvals;
  // loop body stuff
  SCEDA_Graph *dag;
  SCEDA_Vertex *dag_source;
  SCEDA_Vertex *dag_sink;
  GDD_Opcode *dummy_op;
  GDD_NodeLabel *dummy_node;
  SCEDA_List *dag_edges;
  SCEDA_HashMap *ddg_to_body;
  SCEDA_HashMap *longest_path_from_source;
  SCEDA_HashMap *longest_path_to_sink;
} SIRA_Loop;

/** Creates a SIRA Loop.

    @param[in] g = DDG (it is assumed to be lexicographic positive)
    @param[in] IIrec = MII (resource constraints) 

    @return a SIRA Loop. Beware that some values are precomputed using
    the current DDG. If the DDG changes, the values need to be
    updated. */
SIRA_Loop *SIRA_loop_create(SCEDA_Graph *g, int IIrec);

/** Deletes a SIRA Loop.

    @param[in] loop = SIRA Loop to delete */
void SIRA_loop_delete(SIRA_Loop *loop);

/** Updates the precomputed values. This should be called if the DDG
    structure change. 

    @param[in] loop = SIRA Loop 

    @return 0 in case of success, -1 otherwise */
int SIRA_loop_update(SIRA_Loop *loop);

/** Returns the lower bound for scheduling DDG operation u, 
    which must belong to the loop.

    @param[in] loop = SIRA Loop
    @param[in] u = operation

    @return the lower bound */
int SIRA_loop_sigma_min(SIRA_Loop *loop, SCEDA_Vertex *u);

/** Returns the upper bound for scheduling DDG operation u, 
    which must belong to the loop.

    @param[in] loop = SIRA Loop
    @param[in] u = operation
    @param[in] L = duration

    @return the upper bound */
int SIRA_loop_sigma_max(SIRA_Loop *loop, SCEDA_Vertex *u, int L);

/** Returns the lower bound for scheduling killer of DDG operation u,
    which must belong to the loop.

    @param[in] loop = SIRA Loop
    @param[in] u = operation
    @param[in] type = killer type
    @param[in] II = initiation interval

    @return the lower bound */
int SIRA_loop_k_min(SIRA_Loop *loop, SCEDA_Vertex *u, const char *type, int II);

/** Returns the upper bound for scheduling killer of DDG operation u,
    which must belong to the loop.

    @param[in] loop = SIRA Loop
    @param[in] u = operation
    @param[in] L = duration
    @param[in] type = killer type
    @param[in] II = initiation interval

    @return the lower bound */
int SIRA_loop_k_max(SIRA_Loop *loop, SCEDA_Vertex *u, int L, const char *type, int II);

/** Returns the duration for given II (depends on the chosen method).

    @param[in] loop = SIRA Loop
    @param[in] II = II

    @return the duration */
int SIRA_loop_get_L(SIRA_Loop *loop, int II);

/** Set the parameter L used by several methods to compute the max
    latency.

    @param[in] loop = SIRA Loop
    @param[in] L = value for the parameter */
void SIRA_loop_set_L(SIRA_Loop *loop, int L);

/** Switch to lao method for computing max latency.

    @param[in] loop = SIRA Loop
    @param[in] pipelining_degree = pipelining degree */
void SIRA_loop_set_lao_method(SIRA_Loop *loop, int pipelining_degree);

/** Switch to sira method for computing max latency.

    @param[in] loop = SIRA Loop */
void SIRA_loop_set_sira_method(SIRA_Loop *loop);

/** Switch to fixed method for computing max latency.

    @param[in] loop = SIRA Loop */
void SIRA_loop_set_fixed_method(SIRA_Loop *loop);

/** Get the DDG associated to the SIRA Loop.

    @param[in] loop = SIRA Loop
    
    @return the DDG 

    \hideinitializer */
#define SIRA_loop_ddg(loop$) ((loop$)->ddg)

/** Get IImin value (MIIrec) 

    @param[in] loop = SIRA Loop

    @return IImin

    \hideinitializer */
#define SIRA_loop_IImin(loop$) ((loop$)->IImin)

/** Get IImax value (Lmax) 

    @param[in] loop = SIRA Loop

    @return IImax

    \hideinitializer */
#define SIRA_loop_IImax(loop$) ((loop$)->IImax)

/** Get the default duration

    @param[in] loop = SIRA Loop

    @return the default duration

    \hideinitializer */
#define SIRA_loop_Lmax(loop$) ((loop$)->Lmax)

/** Get the maximal distance of the associated DDG.

    @param[in] loop = SIRA Loop

    @return maximal distance

    \hideinitializer */
#define SIRA_loop_max_lambda(loop$) ((loop$)->maxLambda)

#endif

/** \page siraloop Loop: input of SiraLina solver.

    \section siraloop_intro Introduction

    A SIRA Loop is basically a structure based on a DDG which offers
    several facilities to compute some scheduling information.  It is
    used as input value by the siralina solver.

    Given a DDG \f$G = (V,E)\f$ and an execution period \f$p\f$ (also
    called initiation interval II), a periodic scheduling of \f$G\f$
    is a function \f$\sigma: V \to \omega\f$ that satisfies:
    \f$\forall e \in E: \mathop{source}(e) = u \wedge
    \mathop{target}(e) = v \Rightarrow \sigma(u) + \delta(e) \leq
    \sigma(v) + \lambda(e) \times p\f$. We find here that \f$p\f$ must
    be greater than the MII ratio.

    The scheduling of iteration \f$i\f$ of operation \f$u\f$ happens
    at time \f$\sigma(u) + i \times p\f$. Hence the schedule of the
    loop body (which is assumed to be acyclic) defines a periodic
    scheduling of the loop.  

    For computing schedule time bounds, we thus just look at the loop
    body.  We obviously have \f$\underline{\sigma(u)} =
    \mathop{lp}(\top,u) \leq \sigma(u)\f$. Moreover, if \f$L\f$ is an
    upper bound over the total schedule time of the loop body, then we
    have \f$\sigma(u) \leq L - \mathop{lp}(u,\bot) =
    \overline{\sigma(u)}\f$.

    We can also compute bounds over schedule time of killers. We do
    not enter the details here and refer the interested reader to the
    forementioned research articles.

    There exists several ways to compute the duration \f$L\f$.  The
    simplest method is to take a constant value \f$L_G = \sum_{e \in
    E} \delta(e)\f$ but it is also interesting to have a value that
    depends on the desired period, as Sira and Lao method do.

    \section siraloop_api SIRA Loop API

    \code
    SIRA_Loop *SIRA_loop_create(SCEDA_Graph *g, int IIrec);
    \endcode

    Create a SIRA Loop, based on the given DDG, with given resource
    contraints MII.

    \code
    void SIRA_loop_delete(SIRA_Loop *loop);
    \endcode

    Delete a SIRA Loop (but not its associated graph).

    \code
    void SIRA_loop_update(SIRA_Loop *loop);
    \endcode

    Update internal structure of a SIRA Loop, reflecting changes to the
    DDG upon which it is based.

    \code
    int SIRA_loop_sigma_min(SIRA_Loop *loop, SCEDA_Vertex *u);
    int SIRA_loop_sigma_max(SIRA_Loop *loop, SCEDA_Vertex *u, int L);
    int SIRA_loop_k_min(SIRA_Loop *loop, SCEDA_Vertex *u, const char *type, int II);
    int SIRA_loop_k_max(SIRA_Loop *loop, SCEDA_Vertex *u, int L, const char *type, int II);
    \endcode

    Get lower and upper bounds for schedule time of DDG operations or
    killers of a given type.
    
    \code
    void SIRA_loop_set_L(SIRA_Loop *loop, int L);
    void SIRA_loop_set_lao_method(SIRA_Loop *loop, int pipelining_degree);
    void SIRA_loop_set_sira_method(SIRA_Loop *loop);
    void SIRA_loop_set_fixed_method(SIRA_Loop *loop);
    \endcode

    Set method for computing max latency. Set L parameter for these methods.

    \code
    int SIRA_loop_get_L(SIRA_Loop *loop, int II);
    \endcode

    Get max latency for given II, according to current method.

    \code
    SCEDA_Graph *SIRA_loop_ddg(SIRA_Loop *loop);
    \endcode

    Return the DDG upon which is based the SIRA Loop.

    \code
    int SIRA_loop_IImin(SIRA_Loop *loop);
    int SIRA_loop_IImax(SIRA_Loop *loop);
    \endcode

    Return default IImin and IImax values of the SIRA Loop.

    \code
    int SIRA_loop_Lmax(SIRA_Loop *loop);
    \endcode

    Return the default duration of the SIRA Loop, ie return \f$L_G\f$.

    \code
    int SIRA_loop_max_lambda(SIRA_Loop *loop);
    \endcode

    Return the maximal distance of edges in the underlying DDG.

    \section siraloop_exa Example

    \include "siraloop/main.c"

 */
