/*
   Copyright Universite de Versailles Saint-Quentin en Yvelines 2009
   AUTHORS: Sebastien Briais, Sid-Ahmed-Ali Touati

   This file is part of SIRA.
   
   SIRA is free software: you can redistribute it and/or modify it
   under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation, either version 3 of the
   License, or (at your option) any later version.
   
   SIRA is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with SIRA.  If not, see
   <http://www.gnu.org/licenses/>.
*/
#ifndef __SIRA_PARAM_H
#define __SIRA_PARAM_H
/** \file param.h
    \brief SiraLina solver parameters */

/** Type of solver settings. */
typedef struct {
  int ual_flag;
  int bounded_flag;
  int check_flag;
  int max_iter;
} SIRA_Param;

/** Initialise solver settings. 

    \li ual_flag is set to FALSE
    \li bounded_flag is set to TRUE
    \li check_flag is set to TRUE
    \li max_iter is set to 0

    @param[in] settings = solver settings to initialise */
void SIRA_param_init(SIRA_Param *settings);

/** Clean up the solver settings (reclaim local data structures)

    @param[in] settings = solver settings */
void SIRA_param_cleanup(SIRA_Param *settings);

/** Set value of ual flag.
    
    @param[in] settings = solver settings
    @param[in] flag = new value of ual flag

    When set, the ual_flag overrides the architecture description and
    enforces the solver to use UAL semantics. Note that by using a non
    UAL semantics, the resulting DDG might fail to be lexicographic
    positive.

    \hideinitializer */
#define SIRA_param_set_ual_flag(settings$, flag$) ((settings$)->ual_flag = (flag$))

/** Set value of bounded flag.
    
    @param[in] settings = solver settings
    @param[in] flag = new value of bounded flag

    The bounded_flag controls the way schedule time variables (used as
    auxiliary variables by the solver) are bounded. If this flag is
    set, then the bounds are computed using the SIRA_Loop functions.

    \hideinitializer */
#define SIRA_param_set_bounded_flag(settings$, flag$) ((settings$)->bounded_flag = (flag$))

/** Set value of check flag.

    @param[in] settings = solver settings
    @param[in] flag = new value of check flag

    The check flag enforces siralina solver to check that the set of
    reuse edges computed by SiraLina is valid, ie that the associated
    DDG is lexicographic positive. If this is not the case, then it
    will switch to UAL semantics and retry to solve the problem.

    \hideinitializer */
#define SIRA_param_set_check_flag(settings$, flag$) ((settings$)->check_flag = (flag$))

/** Get current value of ual flag.
    
    @param[in] settings = solver settings

    @return current value of ual flag
    \hideinitializer */
#define SIRA_param_get_ual_flag(settings$) ((settings$)->ual_flag)

/** Get current value of bounded flag.
    
    @param[in] settings = solver settings

    @return current value of bounded flag

    \hideinitializer */
#define SIRA_param_get_bounded_flag(settings$) ((settings$)->bounded_flag)

/** Get current value of check flag.

    @param[in] settings = solver settings

    @return current value of check flag

    \hideinitializer */
#define SIRA_param_get_check_flag(settings$) ((settings$)->check_flag)

/** Set value of max_iter
    
    @param[in] settings = solver settings
    @param[in] n = maximal number of iteration used by negative cycles elimination heuristic

    \hideinitializer */
#define SIRA_param_set_max_iter(settings$, n$) ((settings$)->max_iter = (n$))

/** Get value of max_iter
    
    @param[in] settings = solver settings

    @return maximal number of iterations used by negative cycles elimination heuristic
    
    \hideinitializer */
#define SIRA_param_get_max_iter(settings$) ((settings$)->max_iter)

/** \page param Siralina Settings
    
    There exist multiple settings to tweak the exact behaviour of the
    SiraLina solver.  These settings are set through a SIRA_Param
    structure.

    A SIRA_Param structure must be initialised with SIRA_param_init
    function and finalized with SIRA_param_cleanup function.

    We advise to access the SIRA_Param structures exclusively trough
    the functions provided by the API.

    \section param_api API

    \code
    void SIRA_param_init(SIRA_Param *settings);
    \endcode

    Initialise solver settings to default values and create some
    internal data structures.

    \code
    void SIRA_param_cleanup(SIRA_Param *settings);
    \endcode

    Clean up the solver settings. Deallocate internal data structures.

    \code
    int SIRA_param_get_ual_flag(SIRA_Param *settings);
    void SIRA_param_set_ual_flag(SIRA_Param *settings, int flag);
    \endcode

    Get/set value of UAL flag.

    If this flag is set, then all write delay offsets are assumed to
    be equal to 1. It thus changes the formulation (and the solution)
    of the assignment problem.

    \code
    int SIRA_param_get_bounded_flag(SIRA_Param *settings);
    void SIRA_param_set_bounded_flag(SIRA_Param *settings, int flag);
    \endcode
    
    Get/set value of bounded flag.

    If this flag is set, then all variables of the scheduling problem
    are bounded. Bounds are computed thanks to functions defined in
    \ref siraloop.

    \code
    int SIRA_param_get_check_flag(SIRA_Param *settings);
    void SIRA_param_set_check_flag(SIRA_Param *settings, int flag);
    \endcode
    
    Get/set value of check flag.

    If this flag is set, then SiraLina will check that the resulting
    DDG is lexicographic positive. If not, it will retry to solve the
    problem using UAL semantics (the corresponding flag will be modified).

    \code
    int SIRA_param_get_max_iter(SIRA_Param *settings);
    void SIRA_param_set_max_iter(SIRA_Param *settings, int n);
    \endcode
    
    Get/set value of maximal number of iteration when eliminating
    negative cycles.  When set to 0, the iterative heuristic of
    negative cycles elimination is disabled.  

    Note also that depending on the compilation flags, this heuristic
    might be unavailable. In this case, this parameter is ignored.

    When the heuristic of negative cycles elimination is enabled, the
    check flag becomes unuseful since the resulting DDG is guaranteed
    to be lweicographic positive.
*/

#endif

