/*
   Copyright Universite de Versailles Saint-Quentin en Yvelines 2009
   AUTHORS: Sebastien Briais, Sid-Ahmed-Ali Touati

   This file is part of SIRA.
   
   SIRA is free software: you can redistribute it and/or modify it
   under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation, either version 3 of the
   License, or (at your option) any later version.
   
   SIRA is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with SIRA.  If not, see
   <http://www.gnu.org/licenses/>.
*/
#include <stdio.h>

#include <SCEDA/graph_flow.h>
#include <SCEDA/queue.h>
#include <SCEDA/hashmap.h>

#include "solver.h"
#include "util.h"
#include "hungarian/hungarian.h"

/* Include string match and hash functions */
#include "stringutil.h"

#include "boxed.h"

#include "reuse.h"

// Define USE_ILP to use an Integer Linear Program solver (GLPK, LPSolve)
// otherwise use the network flows implementation

#ifdef USE_ILP
// toggle to choose negative cycles elimination heuristic
// by default use SPE heuristic, but RET heuristic is also available
#define SPE_HEURISTIC 

#warning "Using Integer Linear Programming implementation for SiraLina"
// define USE_LPSOLVE to use LPSolve
// otherwise use GLPK
//#define PROBEOBJFUN

#define SIRA_solver_fixed_period_ilp SIRA_solver_fixed_period

#ifdef PROBEOBJFUN
static int obj_fun;
#endif

#ifdef USE_LPSOLVE
#warning "ILP solver is LP Solve"
#include "lp_solve_stub.h"
#else
#warning "ILP solver is GLPK (GPL tainted binary)"
#include "glpk_stub.h"
#endif
#else
#warning "Using Network Flows implementation for SiraLina"
#warning "Iterative heuristic of negative cycles elimination is disabled."
#define SIRA_solver_fixed_period_mcf SIRA_solver_fixed_period
#endif

#define MAX_NAME_SIZE 128

/** Functions to easily manipulate matrix as used by the Hungarian library */
/** Create a squared matrix of desired size */
static int **matrix_create(int size) {
  int **matrix = safe_calloc(size, sizeof(int *));
  int i;
  for(i = 0; i < size; i++) {
    matrix[i] = safe_calloc(size, sizeof(int));
  }
  return matrix;
}

/** Delete a squared matrix of given size */
static void matrix_delete(int size, int **matrix) {
  int i;
  for(i = 0; i < size; i++) {
    free(matrix[i]);
  }
  free(matrix);
}

#ifdef USE_ILP
/** The State data structure helps to keep track of correspondance
    between variables of the linear problem and vertices of the
    graph */
typedef struct {
  int idx; // for name generation
  int sigma; // sigma
  SCEDA_HashMap *sigma_k_map; // sigma of killers (indexed by types)
} State;

static State *state_create() {
  State *st = (State *)safe_malloc(sizeof(State));
  st->sigma_k_map = SCEDA_hashmap_create(NULL, (SCEDA_delete_fun)boxed_delete, (SCEDA_compare_fun)string_match, (SCEDA_hash_fun)string_hash);
  return st;
}

static void state_delete(State *x) {
  SCEDA_hashmap_delete(x->sigma_k_map);
  free(x);
}

/** Initialise the ILP solver 

    @param[in] g = G
    @param[in] weight_map = weight function (types -> boxed(int)) 
    @param[in] states = map from vertices to states

    initialise the states map and return the number of needed
    variables (number of columns)
*/
static int SIRA_solver_init(SCEDA_Graph *g, SIRA_WeightMap *weight_map, SCEDA_HashMap *states) {
  int nb = 0;
  int idx = 0;

  SCEDA_VerticesIterator vertices;
  SCEDA_vertices_iterator_init(g, &vertices);
  while(SCEDA_vertices_iterator_has_next(&vertices)) {
    SCEDA_Vertex *v = SCEDA_vertices_iterator_next(&vertices);
    State *st_v = state_create();
    SCEDA_hashmap_put(states, v, st_v, NULL);
    st_v->idx = idx++;
    st_v->sigma = ++nb;
    
    SIRA_WeightMapIterator weights;
    SIRA_weight_map_iterator_init(weight_map, &weights);
    while(SIRA_weight_map_iterator_has_next(&weights)) {
      char *type;
      int w = SIRA_weight_map_iterator_next(&weights, &type);
      if(w < 0) {
	continue;
      }
      if(GDD_vertex_is_value_of_type(v, type)) {
	SCEDA_hashmap_put(st_v->sigma_k_map, type, boxed_create(int, ++nb), NULL);
      }
    }
    SIRA_weight_map_iterator_cleanup(&weights);
  }
  SCEDA_vertices_iterator_cleanup(&vertices);

  return nb;
}

/** Solve the assignment problem

    @param[in] scheduling_prob = scheduling problem (solved)
    @param[in] vals = V_{R,t}
    @param[in] type = type t
    @param[in] ual_flag = true if UAL semantics
    @param[in] period = scheduling period
    @param[in] states = map V -> State
    @param[in] reuse_edges = list of reuse edges (modified)

    return the cost of the minimal assignment
 */
static int SIRA_solver_assignment_problem(ILP_SOLVER_TYPE *scheduling_prob, 
					  SCEDA_HashSet *vals, const char *type,
					  int ual_flag, int period, 
					  SCEDA_HashMap *states,
					  SCEDA_List *reuse_edges) {
  int n = SCEDA_hashset_size(vals);

  SCEDA_Vertex *values[n];

  {
    int i = 0;
    SCEDA_HashSetIterator vs;
    SCEDA_hashset_iterator_init(vals, &vs);
    while(SCEDA_hashset_iterator_has_next(&vs)) {
      SCEDA_Vertex *v = SCEDA_hashset_iterator_next(&vs);
      values[i++] = v;
    }
    SCEDA_hashset_iterator_cleanup(&vs);
  }

  int **assignment_matrix = matrix_create(n);

  {
    int i;
    for(i = 0; i < n; i++) {
      SCEDA_Vertex *vi = values[i];
      State *st_vi = SCEDA_hashmap_get(states, vi);
      
      boxed(int) sigma_ki_col = SCEDA_hashmap_get(st_vi->sigma_k_map, type);
      
      int sigma_ki = ilp_solver_get_col_value_int(scheduling_prob, boxed_get(sigma_ki_col));
      int j;
      for(j = 0; j < n; j++) {
	SCEDA_Vertex *vj = values[j];
	State *st_vj = SCEDA_hashmap_get(states, vj);
	int sigma_j = ilp_solver_get_col_value_int(scheduling_prob, st_vj->sigma);
	
	int mu_ij = sigma_ki - sigma_j;
	if(ual_flag) {
	  mu_ij++;
	} else {
	  mu_ij -= GDD_vertex_delta_w(vj, type);
	}
	
	mu_ij += period - 1;
	if(mu_ij < 0) {
	  mu_ij = 0;
	}
	mu_ij /= period;
	
	assignment_matrix[i][j] = mu_ij;
      }
    }
  }

  hungarian_problem_t assignment_problem;

  int size = hungarian_init(&assignment_problem, assignment_matrix, n, n, HUNGARIAN_MODE_MINIMIZE_COST);
  assert(size == n);

  hungarian_solve(&assignment_problem);   
  
  int cost = 0;
  {
    int i;
    for(i = 0; i < n; i++) {
      int j;
      for(j = 0; j < n; j++) {
	if(assignment_problem.assignment[i][j] == HUNGARIAN_ASSIGNED) {
	  int mu_ij = assignment_matrix[i][j];
	  cost += mu_ij;
	  SIRA_ReuseEdge *e = SIRA_reuse_edge_create(values[i], values[j], mu_ij);
	  safe_call(SCEDA_list_add(reuse_edges, e));
	}
      }
    }
  }

  hungarian_free(&assignment_problem);

  matrix_delete(n, assignment_matrix);

  return cost;
}

typedef struct {
  int base; // start index 
  int n; // nb values
  SCEDA_Vertex **values; // the values v_1, ..., v_n
  int *reuse_map; // reuse[i] = j is (v_i,v_j) is a reuse edge
  int **mu; // mu_{i,j} 
  int **assignment; // current assignment matrix
} EState;

static int e_state_update(EState *map) {
  hungarian_problem_t assignment_problem;

  int n = map->n;

  int **assignment_matrix = map->assignment;

  int size = hungarian_init(&assignment_problem, assignment_matrix, n, n, HUNGARIAN_MODE_MINIMIZE_COST);
  assert(size == n);
  hungarian_solve(&assignment_problem);

  int has_changed = FALSE;

  {
    int *reuse_map = map->reuse_map;

    int i;
    for(i = 0; i < n; i++) {
      int j;
      for(j = 0; j < n; j++) {
	if(assignment_problem.assignment[i][j] == HUNGARIAN_ASSIGNED) {
/* 	  fprintf(stderr,"%d -> %d",i,j); */
	  if(reuse_map[i] != j) {
/* 	    fprintf(stderr," (changed)"); */
	    has_changed = TRUE;
	    reuse_map[i] = j;
	  }
/* 	  fprintf(stderr,"\n"); */
	}
      }
    }
  }

  hungarian_free(&assignment_problem);

  return has_changed;
}

static EState *e_state_map_create(ILP_SOLVER_TYPE *scheduling_prob,
				  const char *type,
				  SCEDA_HashSet *vals, 
				  int ual_flag, 
				  int period,
				  SCEDA_HashMap *states) {
  EState *map = safe_malloc(sizeof(EState));
  
  int n = SCEDA_hashset_size(vals);
  
  map->n = n;
  
  if(n > 0) {
    SCEDA_Vertex **values = safe_calloc(n, sizeof(SCEDA_Vertex *));
    int *reuse_map = safe_calloc(n, sizeof(int));

    {
      {
	int i;
	SCEDA_HashSetIterator vs;
	SCEDA_hashset_iterator_init(vals, &vs);
	while(SCEDA_hashset_iterator_has_next(&vs)) {
	  SCEDA_Vertex *v = SCEDA_hashset_iterator_next(&vs);
	  reuse_map[i] = i; // initially i reuses i
	  values[i++] = v;
	}
	SCEDA_hashset_iterator_cleanup(&vs);
      }
      
      map->values = values;
      map->reuse_map = reuse_map;
    }
    
    {
      int **assignment_matrix = matrix_create(n);
      
      {
	int i;
	for(i = 0; i < n; i++) {
	  SCEDA_Vertex *vi = values[i];
	  State *st_vi = SCEDA_hashmap_get(states, vi);
	  
	  boxed(int) sigma_ki_col = SCEDA_hashmap_get(st_vi->sigma_k_map, type);
	  
	  int sigma_ki = ilp_solver_get_col_value_int(scheduling_prob, boxed_get(sigma_ki_col));
	  int j;
	  for(j = 0; j < n; j++) {
	    SCEDA_Vertex *vj = values[j];
	    State *st_vj = SCEDA_hashmap_get(states, vj);
	    int sigma_j = ilp_solver_get_col_value_int(scheduling_prob, st_vj->sigma);
	    
	    int mu_ij = sigma_ki - sigma_j;
	    if(ual_flag) {
	      mu_ij++;
	    } else {
	      mu_ij -= GDD_vertex_delta_w(vj, type);
	    }
	    
	    mu_ij += period - 1;
	    if(mu_ij < 0) {
	      mu_ij = 0;
	    }
	    mu_ij /= period;
	    
	    assignment_matrix[i][j] = mu_ij;
	  }
	}
      }
  
      map->mu = assignment_matrix;
    }

    { 
      int **assignment_matrix = matrix_create(n);
      int **mu = map->mu;
      
      {
	int i;
	for(i = 0; i < n; i++) {
	  int j;
	  for(j = 0; j < n; j++) {
	    assignment_matrix[i][j] = mu[i][j];
	  }
	}
      }
      
      map->assignment = assignment_matrix;
    }
  }

  e_state_update(map);

  return map;
}

static void e_state_map_delete(EState *map) {
  int n = map->n;
  if(n > 0) {
    matrix_delete(n, map->mu);
    matrix_delete(n, map->assignment);
    free(map->values);
    free(map->reuse_map);
  }
  free(map);
}

static SCEDA_List *e_state_get_decisions(EState *map) {
  SCEDA_List *reuse_edges = SCEDA_list_create((SCEDA_delete_fun)SIRA_reuse_edge_delete);

  int n = map->n;
  int i;
  int *reuse_map = map->reuse_map;
  int **mu = map->assignment;
  SCEDA_Vertex **values = map->values;
  for(i = 0; i < n; i++) {
    int j = reuse_map[i];
    int mu_ij = mu[i][j];
    SIRA_ReuseEdge *e = SIRA_reuse_edge_create(values[i], values[j], mu_ij);
    safe_call(SCEDA_list_add(reuse_edges, e));
  }
  
  return reuse_edges;
}

#ifdef SPE_HEURISTIC
static void SIRA_solver_enhanced_gen_ddg_constraints_with_killers(ILP_SOLVER_TYPE *negcycle_problem,
								  int offset,
								  SCEDA_Graph *g, 
								  SCEDA_HashMap *states,
								  int N) {
  int ind[1+2];
  ILP_SOLVER_COEF_TYPE coef[1+2];

  // for each edge e=(u,v) add the constraint: x_u - x_v <= lambda(e)
  SCEDA_EdgesIterator edges;
  SCEDA_edges_iterator_init(g, &edges);
  while(SCEDA_edges_iterator_has_next(&edges)) {
    SCEDA_Edge *e = SCEDA_edges_iterator_next(&edges);
    
    SCEDA_Vertex *v_s = SCEDA_edge_source(e);
    SCEDA_Vertex *v_t = SCEDA_edge_target(e);
    
    State *st_s = SCEDA_hashmap_get(states, v_s);
    State *st_t = SCEDA_hashmap_get(states, v_t);
    if(v_s == v_t) {
      ilp_solver_add_row(negcycle_problem, "empty edge", 0, ind, coef, ILP_SOLVER_UP, N * GDD_edge_distance(e) - 1);
    } else {
      ind[1] = offset + st_s->sigma;
      coef[1] = 1;
      ind[2] = offset + st_t->sigma;
      coef[2] = -1;
      ilp_solver_add_row(negcycle_problem, "edge", 2, ind, coef, ILP_SOLVER_UP, N * GDD_edge_distance(e) - 1);
    }
  }
  SCEDA_edges_iterator_cleanup(&edges);

  // for each value u and killer k^t_u of type t
  // for each consumer v of u (of type t)
  // add the constraint x_v - x_{k^t_u} <= - max (lambda(e)) 
  // where e are all the flowdeps edges between u and v
  SCEDA_HashMapIterator vertices;
  SCEDA_hashmap_iterator_init(states, &vertices);
  while(SCEDA_hashmap_iterator_has_next(&vertices)) {
    SCEDA_Vertex *u;
    State *st_u = SCEDA_hashmap_iterator_next(&vertices, &u);
    
    SCEDA_HashMapIterator killers;
    SCEDA_hashmap_iterator_init(st_u->sigma_k_map, &killers);
    while(SCEDA_hashmap_iterator_has_next(&killers)) {
      char *type;
      boxed(int) sigma_k = SCEDA_hashmap_iterator_next(&killers, &type);
      
      SCEDA_HashSet *cons_u = GDD_consumers_of_type(u, type);
      if(SCEDA_hashset_is_empty(cons_u)) {
	ind[1] = offset + boxed_get(sigma_k);
	coef[1] = -1;
	ind[2] = offset + st_u->sigma;
	coef[2] = 1;
	ilp_solver_add_row(negcycle_problem, "empty killing edge", 2, ind, coef, ILP_SOLVER_UP, N * 0 - 1);
      } else {
	SCEDA_HashSetIterator consumers;
	SCEDA_hashset_iterator_init(cons_u, &consumers);
	while(SCEDA_hashset_iterator_has_next(&consumers)) {
	  SCEDA_Vertex *v = SCEDA_hashset_iterator_next(&consumers);
	  State *st_v = SCEDA_hashmap_get(states, v);
	  ind[1] = offset + boxed_get(sigma_k);
	  coef[1] = -1;
	  ind[2] = offset + st_v->sigma;
	  coef[2] = 1;
	  ilp_solver_add_row(negcycle_problem, "killing edge", 2, ind, coef, ILP_SOLVER_UP,N*(-GDD_max_distance_flowdep(u, v, type))-1);
	}
      }
      SCEDA_hashset_delete(cons_u);
    }
    SCEDA_hashmap_iterator_cleanup(&killers);
  }
  SCEDA_hashmap_iterator_cleanup(&vertices);
}

static void SIRA_solver_enhanced_gen_reuse_constraints(ILP_SOLVER_TYPE *negcycle_problem,
						       int offset,
						       SCEDA_HashMap *states,
						       SCEDA_HashMap *e_states,
						       int N) {
  int ind[1+3];
  ILP_SOLVER_COEF_TYPE coef[1+3];
  
  SCEDA_HashMapIterator e_states_iter;
  SCEDA_hashmap_iterator_init(e_states, &e_states_iter);
  while(SCEDA_hashmap_iterator_has_next(&e_states_iter)) {
    char *type;
    EState *map = SCEDA_hashmap_iterator_next(&e_states_iter, &type);
    int **mu = map->assignment;
    int nvalues = map->n;
    SCEDA_Vertex **values = map->values;
    int *reuse = map->reuse_map;
    int i;
    // for each reuse edge e = (u,v) of type t
    // add the constraints x_{k^t_u} - x_v - gamma(e) <= mu(e)
    for(i = 0; i < nvalues; i++) {
      int gamma = 1 + map->base + i;
      SCEDA_Vertex *vi = values[i];
      int j = reuse[i];
      SCEDA_Vertex *vj = values[j];
      State *st_vi = SCEDA_hashmap_get(states, vi);
      State *st_vj = SCEDA_hashmap_get(states, vj);
      boxed(int) sigma_kvi = SCEDA_hashmap_get(st_vi->sigma_k_map, type);
      ind[1] = offset + boxed_get(sigma_kvi);
      coef[1] = 1;
      ind[2] = offset + st_vj->sigma;
      coef[2] = -1;
      ind[3] = gamma;
      coef[3] = -N;
      ilp_solver_add_row(negcycle_problem, "reuse edge", 3, ind, coef, ILP_SOLVER_UP, N * mu[i][j] - 1);
    }
  }
  SCEDA_hashmap_iterator_cleanup(&e_states_iter);
}

#else

static void SIRA_solver_enhanced_gen_ddg_constraints_with_killers(ILP_SOLVER_TYPE *negcycle_problem,
								  int offset,
								  SCEDA_Graph *g, 
								  SCEDA_HashMap *states) {
  int ind[1+2];
  ILP_SOLVER_COEF_TYPE coef[1+2];

  // for each edge e=(u,v) add the constraint: x_u - x_v <= lambda(e)
  SCEDA_EdgesIterator edges;
  SCEDA_edges_iterator_init(g, &edges);
  while(SCEDA_edges_iterator_has_next(&edges)) {
    SCEDA_Edge *e = SCEDA_edges_iterator_next(&edges);
    
    SCEDA_Vertex *v_s = SCEDA_edge_source(e);
    SCEDA_Vertex *v_t = SCEDA_edge_target(e);
    
    State *st_s = SCEDA_hashmap_get(states, v_s);
    State *st_t = SCEDA_hashmap_get(states, v_t);
    if(v_s == v_t) {
      ilp_solver_add_row(negcycle_problem, "empty edge", 0, ind, coef, ILP_SOLVER_UP, GDD_edge_distance(e));
    } else {
      ind[1] = offset + st_s->sigma;
      coef[1] = 1;
      ind[2] = offset + st_t->sigma;
      coef[2] = -1;
      ilp_solver_add_row(negcycle_problem, "edge", 2, ind, coef, ILP_SOLVER_UP, GDD_edge_distance(e));
    }
  }
  SCEDA_edges_iterator_cleanup(&edges);

  // for each value u and killer k^t_u of type t
  // for each consumer v of u (of type t)
  // add the constraint x_v - x_{k^t_u} <= - max (lambda(e)) 
  // where e are all the flowdeps edges between u and v
  SCEDA_HashMapIterator vertices;
  SCEDA_hashmap_iterator_init(states, &vertices);
  while(SCEDA_hashmap_iterator_has_next(&vertices)) {
    SCEDA_Vertex *u;
    State *st_u = SCEDA_hashmap_iterator_next(&vertices, &u);
    
    SCEDA_HashMapIterator killers;
    SCEDA_hashmap_iterator_init(st_u->sigma_k_map, &killers);
    while(SCEDA_hashmap_iterator_has_next(&killers)) {
      char *type;
      boxed(int) sigma_k = SCEDA_hashmap_iterator_next(&killers, &type);
      
      SCEDA_HashSet *cons_u = GDD_consumers_of_type(u, type);
      if(SCEDA_hashset_is_empty(cons_u)) {
	ind[1] = offset + boxed_get(sigma_k);
	coef[1] = -1;
	ind[2] = offset + st_u->sigma;
	coef[2] = 1;
	ilp_solver_add_row(negcycle_problem, "empty killing edge", 2, ind, coef, ILP_SOLVER_UP, 0);
      } else {
	SCEDA_HashSetIterator consumers;
	SCEDA_hashset_iterator_init(cons_u, &consumers);
	while(SCEDA_hashset_iterator_has_next(&consumers)) {
	  SCEDA_Vertex *v = SCEDA_hashset_iterator_next(&consumers);
	  State *st_v = SCEDA_hashmap_get(states, v);
	  ind[1] = offset + boxed_get(sigma_k);
	  coef[1] = -1;
	  ind[2] = offset + st_v->sigma;
	  coef[2] = 1;
	  ilp_solver_add_row(negcycle_problem, "killing edge", 2, ind, coef, ILP_SOLVER_UP,-GDD_max_distance_flowdep(u, v, type));
	}
      }
      SCEDA_hashset_delete(cons_u);
    }
    SCEDA_hashmap_iterator_cleanup(&killers);
  }
  SCEDA_hashmap_iterator_cleanup(&vertices);
}

static void SIRA_solver_enhanced_gen_reuse_constraints(ILP_SOLVER_TYPE *negcycle_problem,
						       int offset,
						       SCEDA_HashMap *states,
						       SCEDA_HashMap *e_states,
						       int gamma_e) {
  int ind[1+3];
  ILP_SOLVER_COEF_TYPE coef[1+3];
  
  SCEDA_HashMapIterator e_states_iter;
  SCEDA_hashmap_iterator_init(e_states, &e_states_iter);
  while(SCEDA_hashmap_iterator_has_next(&e_states_iter)) {
    char *type;
    EState *map = SCEDA_hashmap_iterator_next(&e_states_iter, &type);
    int **mu = map->assignment;
    int nvalues = map->n;
    SCEDA_Vertex **values = map->values;
    int *reuse = map->reuse_map;
    int i;
    // for each reuse edge e = (u,v) of type t
    // add the constraints x_{k^t_u} - x_v - gamma(e) <= mu(e)
    for(i = 0; i < nvalues; i++) {
      int gamma = 1 + map->base + i;
      SCEDA_Vertex *vi = values[i];
      int j = reuse[i];
      SCEDA_Vertex *vj = values[j];
      State *st_vi = SCEDA_hashmap_get(states, vi);
      State *st_vj = SCEDA_hashmap_get(states, vj);
      boxed(int) sigma_kvi = SCEDA_hashmap_get(st_vi->sigma_k_map, type);
      ind[1] = offset + boxed_get(sigma_kvi);
      coef[1] = 1;
      ind[2] = offset + st_vj->sigma;
      coef[2] = -1;
      ind[3] = gamma;
      coef[3] = -1;
      if(gamma == gamma_e) {
	ilp_solver_add_row(negcycle_problem, "reuse edge (t)", 3, ind, coef, ILP_SOLVER_UP, mu[i][j] - 1);
      } else {
	ilp_solver_add_row(negcycle_problem, "reuse edge", 3, ind, coef, ILP_SOLVER_UP, mu[i][j]);
      }
    }
  }
  SCEDA_hashmap_iterator_cleanup(&e_states_iter);
}
#endif

/* #define STATS */

/* #ifdef STATS */
/* int SIRA_max_iter = 0; */
/* #endif */

static int SIRA_solver_enhanced_assignment_problem(SCEDA_Graph *g,
						   ILP_SOLVER_TYPE *scheduling_prob,
						   SIRA_WeightMap *weight_map, 
						   int ual_flag, 
						   int period,
						   SCEDA_HashMap *states,
						   SCEDA_HashMap *reuse_decisions,
						   int max_iter) {
  int n = SCEDA_graph_vcount(g);
  int nv = 0;

/* #ifdef STATS */
/*   SIRA_max_iter = 0; */
/* #endif */

  // map that gives the current enhanced state for each type t
  SCEDA_HashMap *e_states = SCEDA_hashmap_create(NULL, (SCEDA_delete_fun)e_state_map_delete, (SCEDA_compare_fun)string_match, (SCEDA_hash_fun)string_hash);

  // initialisation (solve the first instance of assignment problem for each type)
  { 
    SIRA_WeightMapIterator weights;
    SIRA_weight_map_iterator_init(weight_map, &weights);
    while(SIRA_weight_map_iterator_has_next(&weights)) {
      char *type;
      int w = SIRA_weight_map_iterator_next(&weights, &type);
      if(w < 0) {
	continue;
      }
      SCEDA_HashSet *vals = GDD_get_values_of_type(g, type);
      EState *map = e_state_map_create(scheduling_prob, type, vals, ual_flag, period, states);
      map->base = nv;
      nv += SCEDA_hashset_size(vals);
      SCEDA_hashmap_put(e_states, type, map, NULL);
      SCEDA_hashset_delete(vals);
    }
    SIRA_weight_map_iterator_cleanup(&weights);
  }

  if(!ual_flag) {
    int changed = TRUE;
    int iter;
    for(iter = 0; (iter < max_iter) && changed; iter++) {
      changed = FALSE;

#ifdef SPE_HEURISTIC
      int cols = nv /* gamma */ + (n + nv) /* vertices with killers */;
#else      
      int cols = nv /* gamma */ + nv /* reuse edge */ * (n + nv) /* vertices with killers */;
#endif

/*       fprintf(stderr,"creating problem\n"); */
      ILP_SOLVER_TYPE *negcycle_problem = ilp_solver_create("Neg Cycle Elimination", TRUE, cols);
      
      if(negcycle_problem != NULL) {
	{
/* 	  fprintf(stderr,"generating constraints\n"); */
	  int i;
	  for(i = 1; i <= nv; i++) {
	    ilp_solver_set_col_double(negcycle_problem, i);
	  }
	  
	  for(i = 1 + nv; i <= cols; i++) {
#ifdef SPE_HEURISTIC
	    ilp_solver_set_col_double(negcycle_problem, i);
#endif
	    ilp_solver_set_col_bounds(negcycle_problem, i, 0, 0, 0); // free variables
	  }
	}
	
#ifdef SPE_HEURISTIC
	{ 
	  int offset = nv;
	  int N = n+nv;
	  SIRA_solver_enhanced_gen_ddg_constraints_with_killers(negcycle_problem, offset, g, states, N);
	  SIRA_solver_enhanced_gen_reuse_constraints(negcycle_problem, offset, states, e_states, N);
	}

	SCEDA_HashMapIterator e_states_iter;
	SCEDA_hashmap_iterator_init(e_states, &e_states_iter);
	while(SCEDA_hashmap_iterator_has_next(&e_states_iter)) {
	  char *type;
	  EState *map = SCEDA_hashmap_iterator_next(&e_states_iter, &type);
	  int w = SIRA_weight_map_get_weight_of_type(weight_map, type);
	  
	  int nvalues = map->n;
	  int *reuse = map->reuse_map;
	  int **assignment = map->assignment;
	  int **mu = map->mu;
	  
	  int i;
	  for(i = 0; i < nvalues; i++) {
	    int gamma = 1 + map->base + i;
	    int j = reuse[i];
	    
	    ilp_solver_set_obj_coef(negcycle_problem, gamma, w);
	    
	    ilp_solver_set_col_bounds(negcycle_problem, gamma, ILP_SOLVER_LO, mu[i][j]-assignment[i][j], 0);
	  }
	}
	SCEDA_hashmap_iterator_cleanup(&e_states_iter);
#else
	SCEDA_HashMapIterator e_states_iter;
	SCEDA_hashmap_iterator_init(e_states, &e_states_iter);
	while(SCEDA_hashmap_iterator_has_next(&e_states_iter)) {
	  char *type;
	  EState *map = SCEDA_hashmap_iterator_next(&e_states_iter, &type);
/* 	  fprintf(stderr,"type = %s\n",type); */
	  int w = SIRA_weight_map_get_weight_of_type(weight_map, type);
	  
	  int nvalues = map->n;
	  int *reuse = map->reuse_map;
	  int **assignment = map->assignment;
	  int **mu = map->mu;
	  
	  int i;
	  for(i = 0; i < nvalues; i++) {
	    int offset = nv + (map->base + i) * (n+nv);
	    SIRA_solver_enhanced_gen_ddg_constraints_with_killers(negcycle_problem, offset, g, states);
	    
	    int gamma = 1 + map->base + i;
	    int j = reuse[i];
	    SIRA_solver_enhanced_gen_reuse_constraints(negcycle_problem, offset, states, e_states, gamma);
	    
	    ilp_solver_set_obj_coef(negcycle_problem, gamma, w);
	    
	    ilp_solver_set_col_bounds(negcycle_problem, gamma, ILP_SOLVER_LO, mu[i][j]-assignment[i][j], 0);
	  }
	}
	SCEDA_hashmap_iterator_cleanup(&e_states_iter);
#endif	

/* 	ilp_solver_write_lp(negcycle_problem, "toto"); */
	
/*  	fprintf(stderr,"solving problem\n");  */
	if(ilp_solver_solve(negcycle_problem)) {
/*  	  fprintf(stderr,"problem solved\n");  */
	  
	  SCEDA_HashMapIterator e_states_iter;
	  SCEDA_hashmap_iterator_init(e_states, &e_states_iter);
	  while(SCEDA_hashmap_iterator_has_next(&e_states_iter)) {
	    char *type;
	    EState *map = SCEDA_hashmap_iterator_next(&e_states_iter, &type);
	    
	    int nvalues = map->n;
	    int *reuse = map->reuse_map;
	    int **assignment = map->assignment;
	    
	    int modified = FALSE;
	    
	    int i;
	    for(i = 0; i < nvalues; i++) {
	      int gamma = 1 + map->base + i;
	      int j = reuse[i];
	      double fix = ilp_solver_get_col_value_double(negcycle_problem, gamma);
	      int i_fix = rint(ceil(fix));
/*  	      fprintf(stderr,"reuse(%d,%d,%s) += %d (%g)\n", i, j, type, i_fix, fix);  */
	      if(i_fix != 0) {
		modified = TRUE;
		assignment[i][j] += i_fix;
	      }
	    }
	    
	    if((modified) && (iter < max_iter - 1)) {
/*  	      fprintf(stderr,"assignment %s\n",type);  */
	      changed |= e_state_update(map);
	    }
	  }
	  SCEDA_hashmap_iterator_cleanup(&e_states_iter);
	  
/* 	  if(changed) { */
/* 	    fprintf(stderr,"Assignment changed\n"); */
/* 	  } else { */
/* 	    fprintf(stderr,"Assignment has **not** changed\n"); */
/* 	  } */
	}

/* 	fprintf(stderr,"deleting problem\n"); */
	ilp_solver_delete(negcycle_problem);
      }
    }

/* #ifdef STATS */
/*     SIRA_max_iter = iter; */
/* #endif */
/*     fprintf(stderr,"%d iterations\n",iter); */
  }

/*   fprintf(stderr,"done\n"); */

  // collect reuse decisions
  {
    SCEDA_HashMapIterator decisions;
    SCEDA_hashmap_iterator_init(e_states, &decisions);
    while(SCEDA_hashmap_iterator_has_next(&decisions)) {
      char *type;
      EState *map = SCEDA_hashmap_iterator_next(&decisions, &type);

      SCEDA_List *reuse_edges = e_state_get_decisions(map);

/*       fprintf(stderr,"type %s\n",type); */
      SCEDA_hashmap_put(reuse_decisions, safe_strdup(type), reuse_edges, NULL);
/*       fprintf(stderr,"ok\n"); */
    }
    SCEDA_hashmap_iterator_cleanup(&decisions);
  }

/*   fprintf(stderr,"deleting\n"); */
  // reclaim memory
  SCEDA_hashmap_delete(e_states);

/*   fprintf(stderr,"return\n"); */

  return 0;
}


/** Siralina solver (use ILP solver)

    @param[in] loop = Siraloop DDG
    @param[in] settings = solver settings
    @param[in] period = scheduling period (II)
    @param[in] reuse_decisions = map (type -> reuse edges list) 

    return 0 in case of success, -1 otherwise */
static int SIRA_solver_fixed_period_ilp(SIRA_Loop *loop, 
					SIRA_Param *settings,
					SIRA_WeightMap *weight_map,
					int period,
					SCEDA_HashMap *reuse_decisions) {
  int ual_flag = SIRA_param_get_ual_flag(settings);
  int bounded_flag = SIRA_param_get_bounded_flag(settings);

  int result = -1; // error

  int Lmax = 0;
  if(bounded_flag) {
    Lmax = SIRA_loop_get_L(loop, period);
  }

  SCEDA_Graph *g = SIRA_loop_ddg(loop);

  char namebuf[MAX_NAME_SIZE];

  ilp_solver_init();

  // compute number of columns
  SCEDA_HashMap *states = SCEDA_vertex_map_create((SCEDA_delete_fun)state_delete);

  int cols = SIRA_solver_init(g, weight_map, states);

  ILP_SOLVER_TYPE *scheduling_prob = ilp_solver_create("Scheduling Problem", TRUE, cols);
  if(scheduling_prob == NULL) {
    fprintf(stderr,"SIRA: error while creating linear problem\n");
    SCEDA_hashmap_delete(states);
    ilp_solver_cleanup();
    return -1;
  }

  // set variables names, bounds & objective function
  {
    SCEDA_HashMapIterator vertices;
    SCEDA_hashmap_iterator_init(states, &vertices);
    while(SCEDA_hashmap_iterator_has_next(&vertices)) {
      SCEDA_Vertex *v;
      State *st_v = SCEDA_hashmap_iterator_next(&vertices, &v);

      //      fprintf(stdout,"%s\n", GDD_vertex_mnemonic(v));

      // set name
      snprintf(namebuf, MAX_NAME_SIZE, "sigma(%d)", st_v->idx);
      ilp_solver_set_col_name(scheduling_prob, st_v->sigma, namebuf);

      // set bounds
      if(bounded_flag) {
	int s_min = SIRA_loop_sigma_min(loop, v);
	int s_max = SIRA_loop_sigma_max(loop, v, Lmax);
	ilp_solver_set_col_bounds(scheduling_prob, st_v->sigma, ILP_SOLVER_DB, s_min, s_max);
      } else {
	ilp_solver_set_col_bounds(scheduling_prob, st_v->sigma, ILP_SOLVER_LO, 0, 0); // sigma_i >= 0
      }

      // set objective function coef
      {
	double coef = 0;
	
	SCEDA_HashMapIterator sigma_ks;
	SCEDA_hashmap_iterator_init(st_v->sigma_k_map, &sigma_ks);
	while(SCEDA_hashmap_iterator_has_next(&sigma_ks)) {
	  char *type;
	  boxed(int) sigma_k = SCEDA_hashmap_iterator_next(&sigma_ks, &type);
	  
	  // set name
	  snprintf(namebuf, MAX_NAME_SIZE, "sigma_k(%d,%s)", st_v->idx, type);
	  ilp_solver_set_col_name(scheduling_prob, boxed_get(sigma_k), namebuf);
	  
	  // set bounds
	  if(bounded_flag) {
	    int k_min = SIRA_loop_k_min(loop, v, type, period);
	    int k_max = SIRA_loop_k_max(loop, v, Lmax, type, period);
	    ilp_solver_set_col_bounds(scheduling_prob, boxed_get(sigma_k), ILP_SOLVER_DB, k_min, k_max); // sigmak_i >= 0
	  } else {
	    ilp_solver_set_col_bounds(scheduling_prob, boxed_get(sigma_k), ILP_SOLVER_LO, 0, 0); // sigmak_i >= 0
	  }
	  
	  // set objective function coef
	  int alpha = SIRA_weight_map_get_weight_of_type(weight_map, type);
	  
	  ilp_solver_set_obj_coef(scheduling_prob, boxed_get(sigma_k), alpha);
	  
	  coef += alpha;
	}
	SCEDA_hashmap_iterator_cleanup(&sigma_ks);
	
	// set objective function coef
	ilp_solver_set_obj_coef(scheduling_prob, st_v->sigma, -coef);
      }
    }
    SCEDA_hashmap_iterator_cleanup(&vertices);
  }

  // set linear constraints
  {
    int ind[1+2];
    ILP_SOLVER_COEF_TYPE coef[1+2];

    SCEDA_EdgesIterator edges;
    SCEDA_edges_iterator_init(g, &edges);
    while(SCEDA_edges_iterator_has_next(&edges)) {
      SCEDA_Edge *e = SCEDA_edges_iterator_next(&edges);

      SCEDA_Vertex *v_s = SCEDA_edge_source(e);
      SCEDA_Vertex *v_t = SCEDA_edge_target(e);
      
      State *st_s = SCEDA_hashmap_get(states, v_s);
      State *st_t = SCEDA_hashmap_get(states, v_t);

      snprintf(namebuf, MAX_NAME_SIZE, "sigma(%d)-sigma(%d)", st_t->idx, st_s->idx);
      int lowb = GDD_edge_latency(e) - period * GDD_edge_distance(e);
      if(v_s == v_t) {
	ilp_solver_add_row(scheduling_prob, namebuf, 0, ind, coef, ILP_SOLVER_LO, lowb);
      } else {
	ind[1] = st_s->sigma;
	coef[1] = -1;
	ind[2] = st_t->sigma;
	coef[2] = 1;
	ilp_solver_add_row(scheduling_prob, namebuf, 2, ind, coef, ILP_SOLVER_LO, lowb);
      }
    }
    SCEDA_edges_iterator_cleanup(&edges);

    SCEDA_HashMapIterator vertices;
    SCEDA_hashmap_iterator_init(states, &vertices);
    while(SCEDA_hashmap_iterator_has_next(&vertices)) {
      SCEDA_Vertex *u;
      State *st_u = SCEDA_hashmap_iterator_next(&vertices, &u);

      SCEDA_HashMapIterator sigma_ks;
      SCEDA_hashmap_iterator_init(st_u->sigma_k_map, &sigma_ks);
      while(SCEDA_hashmap_iterator_has_next(&sigma_ks)) {
	char *type;
	boxed(int) sigma_k = SCEDA_hashmap_iterator_next(&sigma_ks, &type);
	
	SCEDA_HashSet *cons_u = GDD_consumers_of_type(u, type);
	if(SCEDA_hashset_is_empty(cons_u)) {
	  snprintf(namebuf, MAX_NAME_SIZE, "sigma_k(%d,%s)", st_u->idx, type);
	  ind[1] = boxed_get(sigma_k);
	  coef[1] = 1;
	  ind[2] = st_u->sigma;
	  coef[2] = -1;
	  int lowb = 1;
	  ilp_solver_add_row(scheduling_prob, namebuf, 2, ind, coef, ILP_SOLVER_LO, lowb);
	} else {
	  SCEDA_HashSetIterator consumers;
	  SCEDA_hashset_iterator_init(cons_u, &consumers);
	  while(SCEDA_hashset_iterator_has_next(&consumers)) {
	    SCEDA_Vertex *v = SCEDA_hashset_iterator_next(&consumers);
	    State *st_v = SCEDA_hashmap_get(states, v);

	    snprintf(namebuf, MAX_NAME_SIZE, "sigma_k(%d,%s)-sigma(%d)", st_u->idx, type, st_v->idx);
	    ind[1] = boxed_get(sigma_k);
	    coef[1] = 1;
	    ind[2] = st_v->sigma;
	    coef[2] = -1;
	    int lowb = GDD_vertex_delta_r(v, type) + period * GDD_max_distance_flowdep(u, v, type);
	    ilp_solver_add_row(scheduling_prob, namebuf, 2, ind, coef, ILP_SOLVER_LO, lowb);
	  }
	  SCEDA_hashset_iterator_cleanup(&consumers);
	}
	SCEDA_hashset_delete(cons_u);
      }
      SCEDA_hashmap_iterator_cleanup(&sigma_ks);
    }
    SCEDA_hashmap_iterator_cleanup(&vertices);
  }

  if(ilp_solver_solve(scheduling_prob)) {
#ifdef PROBEOBJFUN
    obj_fun = rint(ilp_solver_get_obj_value(scheduling_prob));
#endif

#if(0)
    SIRA_WeightMapIterator weights;
    SIRA_weight_map_iterator_init(weight_map, &weights);
    while(SIRA_weight_map_iterator_has_next(&weights)) {
      char *type;
      int w = SIRA_weight_map_iterator_next(&weights, &type);
      if(w < 0) {
	continue;
      }

      SCEDA_HashSet *vals = GDD_get_values_of_type(g, type);
      SCEDA_List *reuse_edges = SCEDA_list_create((SCEDA_delete_fun)SIRA_reuse_edge_delete);

      SIRA_solver_assignment_problem(scheduling_prob, vals, type, ual_flag, period, states, reuse_edges);

      SCEDA_hashmap_put(reuse_decisions, safe_strdup(type), reuse_edges, NULL);

      SCEDA_hashset_delete(vals);
    }
    SIRA_weight_map_iterator_cleanup(&weights);
#else
    SIRA_solver_enhanced_assignment_problem(g, scheduling_prob, weight_map, ual_flag, period, states, reuse_decisions, SIRA_param_get_max_iter(settings));
#endif

    result = 0; // all is OK
  }

  ilp_solver_delete(scheduling_prob);

  SCEDA_hashmap_delete(states);

  ilp_solver_cleanup();

  return result;
}

#else

/** Solve the assignment problem

    @param[in] type = type t
    @param[in] killers = map values v of V_{R,t} to their killer k_v (of type t)
    @param[in] sigma = map vertices to their sigma
    @param[in] ual_flag = true if UAL semantics
    @param[in] period = scheduling period
    @param[in] reuse_edges = list of reuse edges (modified)

    return the cost of the minimal assignment
 */
static int SIRA_solver_assignment_problem_mcf(const char *type, SCEDA_HashMap *killers, SCEDA_HashMap *sigma,
					 int ual_flag, int period, 
					 SCEDA_List *reuse_edges) {
  int n = SCEDA_hashmap_size(killers);
  SCEDA_Vertex *values[n];

  {
    int i = 0;
    SCEDA_HashMapIterator vs;
    SCEDA_hashmap_iterator_init(killers, &vs);
    while(SCEDA_hashmap_iterator_has_next(&vs)) {
      SCEDA_Vertex *v;
      /* SCEDA_Vertex *kv = */ SCEDA_hashmap_iterator_next(&vs, &v);
      values[i++] = v;
    }
    SCEDA_hashmap_iterator_cleanup(&vs);
  }

  int **assignment_matrix = matrix_create(n);

  {
    int i;
    for(i = 0; i < n; i++) {
      SCEDA_Vertex *vi = values[i];
      SCEDA_Vertex *kvi = SCEDA_hashmap_get(killers, vi);

      int sigma_ki = *((int *)SCEDA_hashmap_get(sigma, kvi));

      int j;
      for(j = 0; j < n; j++) {
	SCEDA_Vertex *vj = values[j];
	int sigma_j = *((int *)SCEDA_hashmap_get(sigma, vj));
	
	int mu_ij = sigma_ki - sigma_j;
	if(ual_flag) {
	  mu_ij++;
	} else {
	  mu_ij -= GDD_vertex_delta_w(vj, type);
	}
	
	mu_ij += period - 1;
	if(mu_ij < 0) {
	  mu_ij = 0;
	}
	mu_ij /= period;
	
	assignment_matrix[i][j] = mu_ij;
      }
    }
  }

  hungarian_problem_t assignment_problem;

  int size = hungarian_init(&assignment_problem, assignment_matrix, n, n, HUNGARIAN_MODE_MINIMIZE_COST);
  assert(size == n);

  hungarian_solve(&assignment_problem);   
  
  int cost = 0;
  {
    int i;
    for(i = 0; i < n; i++) {
      int j;
      for(j = 0; j < n; j++) {
	if(assignment_problem.assignment[i][j] == HUNGARIAN_ASSIGNED) {
	  int mu_ij = assignment_matrix[i][j];
	  cost += mu_ij;
	  SIRA_ReuseEdge *e = SIRA_reuse_edge_create(values[i], values[j], mu_ij);
	  safe_call(SCEDA_list_add(reuse_edges, e));
	}
      }
    }
  }

  hungarian_free(&assignment_problem);

  matrix_delete(n, assignment_matrix);

  return cost;
}

/** Deduce sigma values and check optimality of flow 

    @param[in] g = DDG 
    @param[in] cost = cost of edges 
    @param[in] flow = min cost flow 

    @return a map giving for each vertex its sigma or NULL if the flow
    is not optimal */
static SCEDA_HashMap *SIRA_solver_deduce_sigma(SCEDA_Graph *g, SCEDA_HashMap *cost, SCEDA_HashMap *flow) {
  SCEDA_HashMap *dist = SCEDA_vertex_map_create((SCEDA_delete_fun)boxed_delete);

  int n = SCEDA_graph_vcount(g);

  SCEDA_HashSet *in_queue = SCEDA_vertex_set_create();
  SCEDA_Queue *queue = SCEDA_queue_create(NULL);
  
  /* Initially, each node is at distance 0 from a virtual source */
  {
    SCEDA_VerticesIterator vertices;
    SCEDA_vertices_iterator_init(g, &vertices);
    while(SCEDA_vertices_iterator_has_next(&vertices)) {
      SCEDA_Vertex *v = SCEDA_vertices_iterator_next(&vertices);
      safe_call(SCEDA_hashmap_put(dist, v, boxed_create(int, 0), NULL));
      safe_call(SCEDA_queue_enqueue(queue, v));
      safe_call(SCEDA_hashset_add(in_queue, v));  
    }
    SCEDA_vertices_iterator_cleanup(&vertices);
  }

  safe_call(SCEDA_queue_enqueue(queue, NULL));

  /* This is an adaptation of Bellman-Ford algorithm, that works on the residual graph */
  int i = 0;
  while((i < n) && (!SCEDA_queue_is_empty(queue))) {
    SCEDA_Vertex *u;
    safe_call(SCEDA_queue_dequeue(queue, (void **)&u));
    if(u == NULL) {
      i++;
      if(i < n) {
	safe_call(SCEDA_queue_enqueue(queue, NULL));
      }
      continue;
    }
    safe_call(SCEDA_hashset_remove(in_queue, (void **)&u));
    
    boxed(int) dist_u = SCEDA_hashmap_get(dist, u);
   
    /* we first iterate over out edges with positive residual capacity */
    SCEDA_OutEdgesIterator out_edges;
    SCEDA_out_edges_iterator_init(u, &out_edges);
    while(SCEDA_out_edges_iterator_has_next(&out_edges)) {
      SCEDA_Edge *e = SCEDA_out_edges_iterator_next(&out_edges);
      /** every edge has unbounded capacity, so is in the residual graph */
      SCEDA_Vertex *v = SCEDA_edge_target(e);
      boxed(int) dist_v = SCEDA_hashmap_get(dist, v);
      int ce = boxed_get(((boxed(int))SCEDA_hashmap_get(cost, e)));
      if(boxed_get(dist_u) + ce < boxed_get(dist_v)) {
	boxed_set(dist_v, boxed_get(dist_u) + ce);
	if(!SCEDA_hashset_contains(in_queue, v)) {
	  safe_call(SCEDA_queue_enqueue(queue, v));
	  safe_call(SCEDA_hashset_add(in_queue, v));
	}
      }
    }
    SCEDA_out_edges_iterator_cleanup(&out_edges);
    
    /* then we iterate over in edges with positive residual capacity (flow) */
    SCEDA_InEdgesIterator in_edges;
    SCEDA_in_edges_iterator_init(u, &in_edges);
    while(SCEDA_in_edges_iterator_has_next(&in_edges)) {
      SCEDA_Edge *e = SCEDA_in_edges_iterator_next(&in_edges);
      boxed(int) fe = SCEDA_hashmap_get(flow, e);
      if(boxed_get(fe) <= 0) {
	continue;
      }
      SCEDA_Vertex *v = SCEDA_edge_source(e);
      boxed(int) dist_v = SCEDA_hashmap_get(dist, v);
      int ce = -boxed_get(((boxed(int))SCEDA_hashmap_get(cost, e)));
      if(boxed_get(dist_u) + ce < boxed_get(dist_v)) {
	boxed_set(dist_v, boxed_get(dist_u) + ce);
	if(!SCEDA_hashset_contains(in_queue, v)) {
	  safe_call(SCEDA_queue_enqueue(queue, v));
	  safe_call(SCEDA_hashset_add(in_queue, v));
	}
      }
    }
    SCEDA_in_edges_iterator_cleanup(&in_edges);
  }

  SCEDA_hashset_delete(in_queue);

  if(!SCEDA_queue_is_empty(queue)) {
    SCEDA_queue_delete(queue);
    SCEDA_hashmap_delete(dist);
    return NULL;
  }

  SCEDA_queue_delete(queue);

  {
    SCEDA_HashMapIterator vertices;
    SCEDA_hashmap_iterator_init(dist, &vertices);
    while(SCEDA_hashmap_iterator_has_next(&vertices)) {
      SCEDA_Vertex *v;
      boxed(int) dist_v = SCEDA_hashmap_iterator_next(&vertices, &v);
      boxed_set(dist_v, -boxed_get(dist_v));
    }
    SCEDA_hashmap_iterator_cleanup(&vertices);
  }

  return dist;
}

/** Siralina solver (min cost flow implementation) 

    @param[in] loop = Siraloop DDG
    @param[in] settings = solver settings
    @param[in] period = scheduling period (II)
    @param[in] reuse_decisions = map (type -> reuse edges list) 

    return 0 in case of success, -1 otherwise */
static int SIRA_solver_fixed_period_mcf(SIRA_Loop *loop,
				   SIRA_Param *settings,
				   SIRA_WeightMap *weight_map,
				   int period,
				   SCEDA_HashMap *reuse_decisions) {
  SCEDA_Graph *g = SIRA_loop_ddg(loop);

  int ual_flag = SIRA_param_get_ual_flag(settings);

  int result = -1; // error

  /* bookkeep added vertices/edges */
  SCEDA_List *Vk = SCEDA_list_create(NULL); 
  SCEDA_List *Ek = SCEDA_list_create(NULL);

  /* For each type t, a map v -> kv, where v is a value of type t and kv its killer */
  SCEDA_HashMap *killers_map = SCEDA_hashmap_create(NULL, (SCEDA_delete_fun)SCEDA_hashmap_delete, (SCEDA_compare_fun)string_match, (SCEDA_hash_fun)string_hash);

  SCEDA_HashMap *supply = SCEDA_vertex_map_create((SCEDA_delete_fun)boxed_delete);
  SCEDA_HashMap *cost = SCEDA_edge_map_create((SCEDA_delete_fun)boxed_delete);
  int bound = 0;

  /* initialise killers_map */
  {
    SIRA_WeightMapIterator weights;
    SIRA_weight_map_iterator_init(weight_map, &weights);
    while(SIRA_weight_map_iterator_has_next(&weights)) {
      char *type;
      int w = SIRA_weight_map_iterator_next(&weights, &type);
      if(w < 0) {
	continue;
      }
      SCEDA_HashMap *killers = SCEDA_vertex_map_create(NULL);
      safe_call(SCEDA_hashmap_put(killers_map, type, killers, NULL));
    }
    SIRA_weight_map_iterator_cleanup(&weights);
  }

  /* We tweak the C type system */
  /* we first fill killers_map as follows: */
  /* For each type t, a map v -> cons(v, t), where v is a value of type t */
  {
    SCEDA_VerticesIterator vertices;
    SCEDA_vertices_iterator_init(g, &vertices);
    while(SCEDA_vertices_iterator_has_next(&vertices)) {
      SCEDA_Vertex *v = SCEDA_vertices_iterator_next(&vertices);

      int v_supply = 0;

      {
	SIRA_WeightMapIterator weights;
	SIRA_weight_map_iterator_init(weight_map, &weights);
	while(SIRA_weight_map_iterator_has_next(&weights)) {
	  char *type;
	  int w = SIRA_weight_map_iterator_next(&weights, &type);
	  if(w < 0) {
	    continue;
	  }
	  if(GDD_vertex_is_value_of_type(v, type)) {
	    SCEDA_HashMap *killers = SCEDA_hashmap_get(killers_map, type);
	    SCEDA_HashSet *cons_v = GDD_consumers_of_type(v, type);
	    safe_call(SCEDA_hashmap_put(killers, v, cons_v, NULL));
	    v_supply += w;
	  }
	}
	SIRA_weight_map_iterator_cleanup(&weights);
      }

      safe_call(SCEDA_hashmap_put(supply, v, boxed_create(int, v_supply), NULL));
      bound += 2*v_supply;
    }
    SCEDA_vertices_iterator_cleanup(&vertices);
  }

  {
    SCEDA_EdgesIterator edges;
    SCEDA_edges_iterator_init(g, &edges);
    while(SCEDA_edges_iterator_has_next(&edges)) {
      SCEDA_Edge *e = SCEDA_edges_iterator_next(&edges);
      int c = GDD_edge_latency(e) - period * GDD_edge_distance(e);
      safe_call(SCEDA_hashmap_put(cost, e, boxed_create(int, -c), NULL));
    }
    SCEDA_edges_iterator_cleanup(&edges);
  }

  {
    SCEDA_HashMapIterator kmap;
    SCEDA_hashmap_iterator_init(killers_map, &kmap);
    while(SCEDA_hashmap_iterator_has_next(&kmap)) {
      char *type;
      SCEDA_HashMap *killers = SCEDA_hashmap_iterator_next(&kmap, &type);

      int w = SIRA_weight_map_get_weight_of_type(weight_map, type);

      SCEDA_HashMapIterator consumers;
      SCEDA_hashmap_iterator_init(killers, &consumers);
      while(SCEDA_hashmap_iterator_has_next(&consumers)) {
	SCEDA_Vertex *u;
	SCEDA_HashSet *cons_u = SCEDA_hashmap_iterator_next(&consumers, &u);

	SCEDA_Vertex *uk = SCEDA_graph_add_vertex(g, NULL);
	safe_call(SCEDA_list_add(Vk, uk));

	safe_call(SCEDA_hashmap_put(supply, uk, boxed_create(int, -w), NULL));

	/* the following line is correct because it does not modify
	   the structure of the container */
	assert(SCEDA_hashmap_put(killers, u, uk, NULL) == 1); 

	if(SCEDA_hashset_is_empty(cons_u)) {
	  SCEDA_Edge *ek = SCEDA_graph_add_edge(g, u, uk, NULL);
	  safe_call(SCEDA_list_add(Ek, ek));
	  int c = 1;
	  safe_call(SCEDA_hashmap_put(cost, ek, boxed_create(int, -c), NULL));
	} else {
	  SCEDA_HashSetIterator cons;
	  SCEDA_hashset_iterator_init(cons_u, &cons);
	  while(SCEDA_hashset_iterator_has_next(&cons)) {
	    SCEDA_Vertex *v = SCEDA_hashset_iterator_next(&cons);
	    SCEDA_Edge *ek = SCEDA_graph_add_edge(g, v, uk, NULL);
	    safe_call(SCEDA_list_add(Ek, ek));
	    
	    /* the following line is correct because neither u nor v is
	       a killer node and thus, the edges from u to v are not fake */
	    int c = GDD_vertex_delta_r(v, type) + period * GDD_max_distance_flowdep(u, v, type); 
	    safe_call(SCEDA_hashmap_put(cost, ek, boxed_create(int, -c), NULL));
	  }
	  SCEDA_hashset_iterator_cleanup(&cons);
	}

	SCEDA_hashset_delete(cons_u);
      }
      SCEDA_hashmap_iterator_cleanup(&consumers);
    }
    SCEDA_hashmap_iterator_cleanup(&kmap);
  }

  SCEDA_HashMap *flow = SCEDA_graph_min_cost_flow(g, NULL, NULL, NULL, &bound, NULL, supply, NULL, cost);
  if(flow != NULL) {
    SCEDA_HashMap *sigma = SIRA_solver_deduce_sigma(g, cost, flow);
    /** We can forget the flow */
    SCEDA_hashmap_delete(flow); 
    if(sigma != NULL) {
      SCEDA_HashMapIterator killers;
      SCEDA_hashmap_iterator_init(killers_map, &killers);
      while(SCEDA_hashmap_iterator_has_next(&killers)) {
	char *type;
	SCEDA_HashMap *kmap = SCEDA_hashmap_iterator_next(&killers, &type);
	
	SCEDA_List *reuse_edges = SCEDA_list_create((SCEDA_delete_fun)SIRA_reuse_edge_delete);
	
	SIRA_solver_assignment_problem_mcf(type, kmap, sigma, ual_flag, period, reuse_edges);
	
	SCEDA_hashmap_put(reuse_decisions, safe_strdup(type), reuse_edges, NULL);
      }
      SCEDA_hashmap_iterator_cleanup(&killers);
      
      SCEDA_hashmap_delete(sigma);
      
      result = 0; // a solution was found
    }
  }

  {
    SCEDA_ListIterator edges;
    SCEDA_list_iterator_init(Ek, &edges);
    while(SCEDA_list_iterator_has_next(&edges)) {
      SCEDA_Edge *ek = SCEDA_list_iterator_next(&edges);

      void *data;
      SCEDA_graph_remove_edge(g, ek, &data);
    }
    SCEDA_list_iterator_cleanup(&edges);

    SCEDA_list_delete(Ek);
  }

  {
    SCEDA_ListIterator vertices;
    SCEDA_list_iterator_init(Vk, &vertices);
    while(SCEDA_list_iterator_has_next(&vertices)) {
      SCEDA_Vertex *vk = SCEDA_list_iterator_next(&vertices);

      void *data;
      SCEDA_graph_remove_vertex(g, vk, &data);
    }
    SCEDA_list_iterator_cleanup(&vertices);

    SCEDA_list_delete(Vk);
  }

  SCEDA_hashmap_delete(killers_map);

  SCEDA_hashmap_delete(supply);
  SCEDA_hashmap_delete(cost);

  return result;
}

#endif

/** Check whether the solution satisfies the register constraints.

    @param[in] regconstraints_map = register constraints
    @param[in] reuse_decisions = map (type -> reuse edges list)

    @return TRUE iff constraints are satisfied */
static int SIRA_solver_is_solution(SIRA_RegConstraintsMap *regconstraints_map, SIRA_WeightMap *weight_map, SCEDA_HashMap *reuse_decisions) {
  int is_solution = TRUE;
  //
  SCEDA_HashMapIterator decisions;
  SCEDA_hashmap_iterator_init(reuse_decisions, &decisions);
  while(SCEDA_hashmap_iterator_has_next(&decisions)) {
    char *type;
    SCEDA_List *reuse_edges = SCEDA_hashmap_iterator_next(&decisions, &type);
    int w = SIRA_weight_map_get_weight_of_type(weight_map, type);
    if(w <= 0) {
      continue;
    }
    int cost = SIRA_reuse_decisions_get_cost(reuse_edges);
    int max = SIRA_regconstraints_map_get_constraint_of_type(regconstraints_map, type);
    if(cost > max) {
      is_solution = FALSE;
      break;
    }
  }
  SCEDA_hashmap_iterator_cleanup(&decisions);
  //
  return is_solution;
}

/** Check whether the solution is valid, i.e. if the resulting DDG is
    lexicographic positive.

    @param[in] loop = SiraLoop DDG
    @param[in] settings = solver settings
    @param[in] reuse_decisions = map (type -> reuse edges list)

    @return TRUE iff resulting DDG is lexicographic positive 
    note that it is trivially TRUE when UAL semantics is enabled */
static int SIRA_solver_solution_is_lexicographic_positive(SIRA_Loop *loop, SIRA_Param *settings, SCEDA_HashMap *reuse_decisions) {
  int result = TRUE;
  if(SIRA_param_get_ual_flag(settings)) {
    return result;
  }

  /* check whether the modified DDG is lexicographic positive */
  /* apply the reuse decisions */
  SCEDA_List *all_antidep_edges = SCEDA_list_create((SCEDA_delete_fun)SCEDA_list_delete);
  {
    SCEDA_HashMapIterator decisions;
    SCEDA_hashmap_iterator_init(reuse_decisions, &decisions);
    while(SCEDA_hashmap_iterator_has_next(&decisions)) {
      char *type;
      SCEDA_List *reuse_edges = SCEDA_hashmap_iterator_next(&decisions, &type);
      SCEDA_List *antidep_edges = SIRA_reuse_apply_decisions(SIRA_loop_ddg(loop), type, reuse_edges, SIRA_param_get_ual_flag(settings));
      safe_call(SCEDA_list_add(all_antidep_edges, antidep_edges));
    }
    SCEDA_hashmap_iterator_cleanup(&decisions);
  }

  /* check lexicographic positiveness */
  SCEDA_List *cycle;
  if(!GDD_is_lexicographic_positive(SIRA_loop_ddg(loop), &cycle)) {
    SCEDA_list_delete(cycle);
    result = FALSE;
  }

  /* unapply reuse decisions */
  {
    SCEDA_ListIterator ae;
    SCEDA_list_iterator_init(all_antidep_edges, &ae);
    while(SCEDA_list_iterator_has_next(&ae)) {
      SCEDA_List *antidep_edges = SCEDA_list_iterator_next(&ae);
      GDD_delete_edges(SIRA_loop_ddg(loop), antidep_edges);
    }
    SCEDA_list_iterator_cleanup(&ae);
  }
  SCEDA_list_delete(all_antidep_edges);
  
  return result;
}

/** Apply siralina heuristic for every II between IImin and IImax and
    print the results (used for statistics).

    @param[in] loop = siraloop
    @param[in] settings = solver settings
    @param[in] IImin = IImin (if negative, IImin of the loop will be used)
    @param[in] IImax = IImax (if negative, IImax of the loop will be used) */
void SIRA_solver_stats(SIRA_Loop *loop, SIRA_Param *settings, SIRA_WeightMap *weight_map, int IImin, int IImax) {
  if(IImin < 0) {
    IImin = SIRA_loop_IImin(loop);
  }
  if(IImax < 0) {
    IImax = SIRA_loop_IImax(loop);
  }

  SCEDA_HashMap *reuse_decisions = SCEDA_hashmap_create((SCEDA_delete_fun)free, (SCEDA_delete_fun)SCEDA_list_delete, (SCEDA_match_fun)string_match, (SCEDA_hash_fun)string_hash);

  int II;
  for(II = IImin; II <= IImax; II++) {
    if(SIRA_solver_fixed_period(loop, settings, weight_map, II, reuse_decisions) == 0) {
      fprintf(stdout,"II= %d\tL= %d\t", II, SIRA_loop_get_L(loop, II));

      SCEDA_HashMapIterator decisions;
      SCEDA_hashmap_iterator_init(reuse_decisions, &decisions);
      while(SCEDA_hashmap_iterator_has_next(&decisions)) {
	char *type;
	SCEDA_List *reuse_edges = SCEDA_hashmap_iterator_next(&decisions, &type);
	int cost = SIRA_reuse_decisions_get_cost(reuse_edges);
	fprintf(stdout, "R_%s= %d\t", type, cost);
      }
      SCEDA_hashmap_iterator_cleanup(&decisions);

#ifdef PROBEOBJFUN
      fprintf(stdout,"obj= %d\t", obj_fun);
#endif

      fprintf(stdout,"\n");
    }
    SCEDA_hashmap_clear(reuse_decisions);
  }
  
  SCEDA_hashmap_delete(reuse_decisions);
}

/** Apply siralina heuristic for given II.

    @param[in] loop = siraloop
    @param[in] settings = solver settings
    @param[in] II = II (if negative, fails)
    @param[out] solution = computed solution (map from types to reuse edges list)

    @return 0 if a solution was found, otherwise -1 */
static int SIRA_solver_single_internal(SIRA_Loop *loop, SIRA_Param *settings, SIRA_WeightMap *weight_map, int II, SCEDA_HashMap **solution) {
  if(II < 0) {
    return -1;
  }

  SCEDA_HashMap *reuse_decisions = SCEDA_hashmap_create((SCEDA_delete_fun)free, (SCEDA_delete_fun)SCEDA_list_delete, (SCEDA_match_fun)string_match, (SCEDA_hash_fun)string_hash);

  if(SIRA_solver_fixed_period(loop, settings, weight_map, II, reuse_decisions) == 0) {
    *solution = reuse_decisions;
    return 0;
  } else {
    SCEDA_hashmap_delete(reuse_decisions);
    return -1;
  }
}

/** Apply siralina heuristic for given II.

    @param[in] loop = siraloop
    @param[in] settings = solver settings (semantics might be modified)
    @param[in] II = II (if negative, fails)
    @param[out] solution = computed solution (map from types to reuse edges list)

    @return 0 if a solution was found, otherwise -1 */
int SIRA_solver_single(SIRA_Loop *loop, SIRA_Param *settings, SIRA_WeightMap *weight_map, int II, SCEDA_HashMap **solution) {
  if(SIRA_solver_single_internal(loop, settings, weight_map, II, solution) == 0) {
    /* Do we have to check the result ? */
    if(!SIRA_param_get_check_flag(settings)) {
      return 0;
    }
    if(SIRA_param_get_ual_flag(settings)) {
      /* if UAL semantics then no need to check solution */
      return 0;
    } else {
      if(SIRA_solver_solution_is_lexicographic_positive(loop, settings, *solution)) {
	return 0;
      } 
      /* the solution was not lexicographic positive */
      /* retry with UAL semantics */
      SCEDA_hashmap_delete(*solution);
      SIRA_param_set_ual_flag(settings, TRUE);
      return SIRA_solver_single_internal(loop, settings, weight_map, II, solution);
    }
  }
  return -1;
}

/** Apply siralina heuristic for every II between IImin and IImax and
    stops when a solution is found (ie satisfies the register
    constraints).

    Before beginning the search procedure, a simple heuristic is
    implemented to determine whether there is a solution to the
    problem.

    @param[in] loop = siraloop
    @param[in] regconstraints_map = register contraints
    @param[in] settings = solver settings
    @param[in] IImin = IImin (if negative, IImin of the loop will be used)
    @param[in] IImax = IImax (if negative, IImax of the loop will be used) 
    @param[out] solution = computed solution (map from types to reuse edges list)

    @return II >= 0 if a solution was found, otherwise -1 */
static int SIRA_solver_search_internal(SIRA_Loop *loop, SIRA_RegConstraintsMap *regconstraints_map, SIRA_Param *settings, SIRA_WeightMap *weight_map, int IImin, int IImax, SCEDA_HashMap **solution) {
  int found = FALSE;

  if(IImin < 0) {
    IImin = SIRA_loop_IImin(loop);
  }
  if(IImax < 0) {
    IImax = SIRA_loop_IImax(loop);
  }
  if(IImax < 0) {
    return -1;
  }
  if(IImax < IImin) {
    return -1;
  }

  SCEDA_HashMap *reuse_decisions = SCEDA_hashmap_create((SCEDA_delete_fun)free, (SCEDA_delete_fun)SCEDA_list_delete, (SCEDA_match_fun)string_match, (SCEDA_hash_fun)string_hash);

  int II = IImin;
  SCEDA_hashmap_clear(reuse_decisions);
  if(SIRA_solver_fixed_period(loop, settings, weight_map, II, reuse_decisions) == 0) {
    found = SIRA_solver_is_solution(regconstraints_map, weight_map, reuse_decisions);
  }

  if(!found) {
    II = IImax;
    SCEDA_hashmap_clear(reuse_decisions);
    if(SIRA_solver_fixed_period(loop, settings, weight_map, II, reuse_decisions) == 0) {
      found = SIRA_solver_is_solution(regconstraints_map, weight_map, reuse_decisions);
    }

    if(found) {
      found = FALSE;
      for(II = IImin+1; II <= IImax; II++) {
	SCEDA_hashmap_clear(reuse_decisions);
	if(SIRA_solver_fixed_period(loop, settings, weight_map, II, reuse_decisions) == 0) {
	  found = SIRA_solver_is_solution(regconstraints_map, weight_map, reuse_decisions);
	}
	if(found) {
	  break;
	}
      }
    }
  }

  if(found) {
    *solution = reuse_decisions;
    return II;
  } else {
    SCEDA_hashmap_delete(reuse_decisions);
    return -1;
  }
}

/** Apply siralina heuristic for every II between IImin and IImax and
    stops when a solution is found (ie satisfies the register
    constraints).

    Before beginning the search procedure, a simple heuristic is
    implemented to determine whether there is a solution to the
    problem.

    @param[in] loop = siraloop
    @param[in] regconstraints_map = register contraints
    @param[in] settings = solver settings (semantics might be modified)
    @param[in] IImin = IImin (if negative, IImin of the loop will be used)
    @param[in] IImax = IImax (if negative, IImax of the loop will be used) 
    @param[out] solution = computed solution (map from types to reuse edges list)

    @return II >= 0 if a solution was found, otherwise -1 */
int SIRA_solver_search(SIRA_Loop *loop, SIRA_RegConstraintsMap *regconstraints_map, SIRA_Param *settings, SIRA_WeightMap *weight_map, int IImin, int IImax, SCEDA_HashMap **solution) {
  int II = SIRA_solver_search_internal(loop, regconstraints_map, settings, weight_map, IImin, IImax, solution);
  if(II < 0) {
    return II;
  }
  /** Do we have to check the result ? */
  if(!SIRA_param_get_check_flag(settings)) {
    return II;
  }
  /** If UAL semantics, no need to check */
  if(SIRA_param_get_ual_flag(settings)) {
    return II;
  }
  if(SIRA_solver_solution_is_lexicographic_positive(loop, settings, *solution)) {
    return II;
  } 
  /* the solution was not lexicographic positive */
  /* retry with UAL semantics */
  SCEDA_hashmap_delete(*solution);
  SIRA_param_set_ual_flag(settings, TRUE);
  return SIRA_solver_search_internal(loop, regconstraints_map, settings, weight_map, IImin, IImax, solution);
}

